#!/usr/bin/env python3
#
#  actinides.py
r"""
Actinides (or actinoids) in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: Ac

	:class:`~chemistry_tools.elements.classes.Element` representing Actinium

.. data:: Th

	:class:`~chemistry_tools.elements.classes.Element` representing Thorium

.. data:: Pa

	:class:`~chemistry_tools.elements.classes.Element` representing Protactinium

.. data:: U

	:class:`~chemistry_tools.elements.classes.Element` representing Uranium

.. data:: Np

	:class:`~chemistry_tools.elements.classes.Element` representing Neptunium

.. data:: Pu

	:class:`~chemistry_tools.elements.classes.Element` representing Plutonium

.. data:: Am

	:class:`~chemistry_tools.elements.classes.Element` representing Americium

.. data:: Cm

	:class:`~chemistry_tools.elements.classes.Element` representing Curium

.. data:: Bk

	:class:`~chemistry_tools.elements.classes.Element` representing Berkelium

.. data:: Cf

	:class:`~chemistry_tools.elements.classes.Element` representing Californium

.. data:: Es

	:class:`~chemistry_tools.elements.classes.Element` representing Einsteinium

.. data:: Fm

	:class:`~chemistry_tools.elements.classes.Element` representing Fermium

.. data:: Md

	:class:`~chemistry_tools.elements.classes.Element` representing Mendelevium

.. data:: No

	:class:`~chemistry_tools.elements.classes.Element` representing Nobelium

.. data:: Lr

	:class:`~chemistry_tools.elements.classes.Element` representing Lawrencium

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

# this package
from . import classes

__all__ = ("Ac", "Th", "Pa", 'U', "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr")

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

Element = classes.Element

Ac = Element(
		89,
		"Ac",
		"Actinium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=227.0278,
		eleneg=1.1,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=3470.0,
		tmelt=1324.0,
		density=10.07,
		eleconfig="[Rn] 6d 7s2",
		oxistates="3*",
		ionenergy=(5.17, 12.1),
		isotopes={
				206: (206.0145, 0.0),
				207: (207.01195, 0.0),
				208: (208.01155, 0.0),
				209: (209.00949, 0.0),
				210: (210.00944, 0.0),
				211: (211.00773, 0.0),
				212: (212.00781, 0.0),
				213: (213.00661, 0.0),
				214: (214.006902, 0.0),
				215: (215.006454, 0.0),
				216: (216.00872, 0.0),
				217: (217.009347, 0.0),
				218: (218.01164, 0.0),
				219: (219.01242, 0.0),
				220: (220.014763, 0.0),
				221: (221.01559, 0.0),
				222: (222.017844, 0.0),
				223: (223.019137, 0.0),
				224: (224.021723, 0.0),
				225: (225.02323, 0.0),
				226: (226.026098, 0.0),
				227: (227.0277521, 0.0),
				228: (228.0310211, 0.0),
				229: (229.03302, 0.0),
				230: (230.03629, 0.0),
				231: (231.03856, 0.0),
				232: (232.04203, 0.0),
				233: (233.04455, 0.0),
				234: (234.04842, 0.0),
				235: (235.05123, 0.0),
				236: (236.0553, 0.0),
				},
		description=(
				"Silvery radioactive metallic element, belongs to group 3 of the "
				"periodic table. The most stable isotope, Ac-227, has a half-life of "
				"217 years. Ac-228 (half-life of 6.13 hours) also occurs in nature. "
				"There are 22 other artificial isotopes, all radioactive and having "
				"very short half-lives. Chemistry similar to lanthanumpy. Used as a "
				"source of alpha particles. Discovered by A. Debierne in 1899."
				)
		)

Th = Element(
		90,
		"Th",
		"Thorium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=232.0377,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=1.65,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=5060.0,
		tmelt=2028.0,
		density=11.72,
		eleconfig="[Rn] 6d2 7s2",
		oxistates="4*",
		ionenergy=(6.3067, 11.5, 20.0, 28.8),
		isotopes={
				209: (209.01772, 0.0),
				210: (210.015075, 0.0),
				211: (211.01493, 0.0),
				212: (212.01298, 0.0),
				213: (213.01301, 0.0),
				214: (214.0115, 0.0),
				215: (215.01173, 0.0),
				216: (216.011062, 0.0),
				217: (217.013114, 0.0),
				218: (218.013284, 0.0),
				219: (219.01554, 0.0),
				220: (220.015748, 0.0),
				221: (221.018184, 0.0),
				222: (222.018468, 0.0),
				223: (223.020811, 0.0),
				224: (224.021467, 0.0),
				225: (225.023951, 0.0),
				226: (226.024903, 0.0),
				227: (227.0277041, 0.0),
				228: (228.0287411, 0.0),
				229: (229.031762, 0.0),
				230: (230.0331338, 0.0),
				231: (231.0363043, 0.0),
				232: (232.0380553, 1.0),
				233: (233.0415818, 0.0),
				234: (234.043601, 0.0),
				235: (235.04751, 0.0),
				236: (236.04987, 0.0),
				237: (237.05389, 0.0),
				238: (238.0565, 0.0),
				},
		description=(
				"Grey radioactive metallic element. Belongs to actinoids. Found in "
				"monazite sand in Brazil, India and the US. Thorium-232 has a "
				"half-life of 1.39x10^10 years. Can be used as a nuclear fuel for "
				"breeder reactors. Thorium-232 captures slow Neutrons and breeds "
				"uranium-233. Discovered by Jons J. Berzelius in 1829."
				)
		)

Pa = Element(
		91,
		"Pa",
		"Protactinium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=231.03588,
		eleneg=1.5,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=4300.0,
		tmelt=1845.0,
		density=15.37,
		eleconfig="[Rn] 5f2 6d 7s2",
		oxistates="5*, 4",
		ionenergy=(5.89, ),
		isotopes={
				212: (212.0232, 0.0),
				213: (213.02111, 0.0),
				214: (214.02092, 0.0),
				215: (215.01919, 0.0),
				216: (216.01911, 0.0),
				217: (217.01832, 0.0),
				218: (218.020042, 0.0),
				219: (219.01988, 0.0),
				220: (220.02188, 0.0),
				221: (221.02188, 0.0),
				222: (222.02374, 0.0),
				223: (223.02396, 0.0),
				224: (224.025626, 0.0),
				225: (225.02613, 0.0),
				226: (226.027948, 0.0),
				227: (227.028805, 0.0),
				228: (228.031051, 0.0),
				229: (229.0320968, 0.0),
				230: (230.034541, 0.0),
				231: (231.035884, 1.0),
				232: (232.038592, 0.0),
				233: (233.0402473, 0.0),
				234: (234.043308, 0.0),
				235: (235.04544, 0.0),
				236: (236.04868, 0.0),
				237: (237.05115, 0.0),
				238: (238.0545, 0.0),
				239: (239.05726, 0.0),
				240: (240.06098, 0.0),
				},
		description=(
				"Radioactive metallic element, belongs to the actinoids. The most "
				"stable isotope, Pa-231 has a half-life of 2.43*10^4 years. At least "
				"10 other radioactive isotopes are known. No practical applications "
				"are known. Discovered in 1917 by Lise Meitner and Otto Hahn."
				)
		)
U = Element(
		92,
		'U',
		"Uranium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=238.02891,
		eleneg=1.38,
		eleaffin=0.0,
		covrad=1.42,
		atmrad=0.0,
		vdwrad=1.86,
		tboil=4407.0,
		tmelt=1408.0,
		density=18.97,
		eleconfig="[Rn] 5f3 6d 7s2",
		oxistates="6*, 5, 4, 3",
		ionenergy=(6.1941, ),
		isotopes={
				217: (217.02437, 0.0),
				218: (218.02354, 0.0),
				219: (219.02492, 0.0),
				220: (220.02472, 0.0),
				221: (221.0264, 0.0),
				222: (222.02609, 0.0),
				223: (223.02774, 0.0),
				224: (224.027605, 0.0),
				225: (225.029391, 0.0),
				226: (226.029339, 0.0),
				227: (227.031156, 0.0),
				228: (228.031374, 0.0),
				229: (229.033506, 0.0),
				230: (230.03394, 0.0),
				231: (231.036294, 0.0),
				232: (232.0371562, 0.0),
				233: (233.0396352, 0.0),
				234: (234.0409521, 5.4e-05),
				235: (235.0439299, 0.007204),
				236: (236.045568, 0.0),
				237: (237.0487302, 0.0),
				238: (238.0507882, 0.992742),
				239: (239.0542933, 0.0),
				240: (240.056592, 0.0),
				241: (241.06033, 0.0),
				242: (242.06293, 0.0),
				},
		description=(
				"White radioactive metallic element belonging to the actinoids. "
				"Three natural isotopes, U-238, U-235 and U-234. Uranium-235 is used "
				"as the fuel for nuclear reactors and weapons. Discovered by Martin "
				"H. Klaproth in 1789."
				)
		)

Np = Element(
		93,
		"Np",
		"Neptunium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=237.0482,
		eleneg=1.36,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=4175.0,
		tmelt=912.0,
		density=20.48,
		eleconfig="[Rn] 5f4 6d 7s2",
		oxistates="6, 5*, 4, 3",
		ionenergy=(6.2657, ),
		isotopes={
				225: (225.03391, 0.0),
				226: (226.03515, 0.0),
				227: (227.03496, 0.0),
				228: (228.03618, 0.0),
				229: (229.03626, 0.0),
				230: (230.03783, 0.0),
				231: (231.03825, 0.0),
				232: (232.04011, 0.0),
				233: (233.04074, 0.0),
				234: (234.042895, 0.0),
				235: (235.0440633, 0.0),
				236: (236.04657, 0.0),
				237: (237.0481734, 0.0),
				238: (238.0509464, 0.0),
				239: (239.052939, 0.0),
				240: (240.056162, 0.0),
				241: (241.05825, 0.0),
				242: (242.06164, 0.0),
				243: (243.06428, 0.0),
				244: (244.06785, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element, belongs to the "
				"actinoids. Np-237, the most stable isotope, has a half-life of "
				"2.2*10^6 years and is a by product of nuclear reactors. The other "
				"known isotopes have mass numbers 229 through 236, and 238 through "
				"241. Np-236 has a half-life of 5*10^3 years. First produced by "
				"Edwin M. McMillan and P.H. Abelson in 1940."
				)
		)

Pu = Element(
		94,
		"Pu",
		"Plutonium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=244.0642,
		eleneg=1.28,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=3505.0,
		tmelt=913.0,
		density=19.74,
		eleconfig="[Rn] 5f6 7s2",
		oxistates="6, 5, 4*, 3",
		ionenergy=(6.026, ),
		isotopes={
				228: (228.03874, 0.0),
				229: (229.04015, 0.0),
				230: (230.03965, 0.0),
				231: (231.041101, 0.0),
				232: (232.041187, 0.0),
				233: (233.043, 0.0),
				234: (234.043317, 0.0),
				235: (235.045286, 0.0),
				236: (236.046058, 0.0),
				237: (237.0484097, 0.0),
				238: (238.0495599, 0.0),
				239: (239.0521634, 0.0),
				240: (240.0538135, 0.0),
				241: (241.0568515, 0.0),
				242: (242.0587426, 0.0),
				243: (243.062003, 0.0),
				244: (244.064204, 0.0),
				245: (245.067747, 0.0),
				246: (246.070205, 0.0),
				247: (247.07407, 0.0),
				},
		description=(
				"Dense silvery radioactive metallic transuranic element, belongs to "
				"the actinoids. Pu-244 is the most stable isotope with a half-life "
				"of 7.6*10^7 years. Thirteen isotopes are known. Pu-239 is the most "
				"important, it undergoes nuclear fission with slow neutrons and is "
				"hence important to nuclear weapons and reactors. Plutonium "
				"production is monitored down to the gram to prevent military "
				"misuse. First produced by Gleen T. Seaborg, Edwin M. McMillan, J.W. "
				"Kennedy and A.C. Wahl in 1940."
				)
		)

Am = Element(
		95,
		"Am",
		"Americium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=243.0614,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=2880.0,
		tmelt=1449.0,
		density=13.67,
		eleconfig="[Rn] 5f7 7s2",
		oxistates="6, 5, 4, 3*",
		ionenergy=(5.9738, ),
		isotopes={
				231: (231.04556, 0.0),
				232: (232.04659, 0.0),
				233: (233.04635, 0.0),
				234: (234.04781, 0.0),
				235: (235.04795, 0.0),
				236: (236.04958, 0.0),
				237: (237.05, 0.0),
				238: (238.05198, 0.0),
				239: (239.0530245, 0.0),
				240: (240.0553, 0.0),
				241: (241.0568291, 0.0),
				242: (242.0595492, 0.0),
				243: (243.0613811, 0.0),
				244: (244.0642848, 0.0),
				245: (245.066452, 0.0),
				246: (246.069775, 0.0),
				247: (247.07209, 0.0),
				248: (248.07575, 0.0),
				249: (249.07848, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element, belongs to the "
				"actinoids. Ten known isotopes. Am-243 is the most stable isotope, "
				"with a half-life of 7.95*10^3 years. Discovered by Glenn T. Seaborg "
				"and associates in 1945, it was obtained by bombarding Uranium-238 "
				"with alpha particles."
				)
		)

Cm = Element(
		96,
		"Cm",
		"Curium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=247.0704,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=1620.0,
		density=13.51,
		eleconfig="[Rn] 5f7 6d 7s2",
		oxistates="4, 3*",
		ionenergy=(5.9914, ),
		isotopes={
				233: (233.05077, 0.0),
				234: (234.05016, 0.0),
				235: (235.05143, 0.0),
				236: (236.05141, 0.0),
				237: (237.0529, 0.0),
				238: (238.05303, 0.0),
				239: (239.05496, 0.0),
				240: (240.0555295, 0.0),
				241: (241.057653, 0.0),
				242: (242.0588358, 0.0),
				243: (243.0613891, 0.0),
				244: (244.0627526, 0.0),
				245: (245.0654912, 0.0),
				246: (246.0672237, 0.0),
				247: (247.070354, 0.0),
				248: (248.072349, 0.0),
				249: (249.075953, 0.0),
				250: (250.078357, 0.0),
				251: (251.082285, 0.0),
				252: (252.08487, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element. Belongs to actinoid "
				"series. Nine known isotopes, Cm-247 has a half-life of 1.64*10^7 "
				"years. First identified by Glenn T. Seaborg and associates in 1944, "
				"first produced by L.B. Werner and I. Perlman in 1947 by bombarding "
				"americium-241 with Neutrons. Named for Marie Curie."
				)
		)

Bk = Element(
		97,
		"Bk",
		"Berkelium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=247.0703,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=1258.0,
		density=13.25,
		eleconfig="[Rn] 5f9 7s2",
		oxistates="4, 3*",
		ionenergy=(6.1979, ),
		isotopes={
				235: (235.05658, 0.0),
				236: (236.05733, 0.0),
				237: (237.057, 0.0),
				238: (238.05828, 0.0),
				239: (239.05828, 0.0),
				240: (240.05976, 0.0),
				241: (241.06023, 0.0),
				242: (242.06198, 0.0),
				243: (243.063008, 0.0),
				244: (244.065181, 0.0),
				245: (245.0663616, 0.0),
				246: (246.06867, 0.0),
				247: (247.070307, 0.0),
				248: (248.07309, 0.0),
				249: (249.0749867, 0.0),
				250: (250.078317, 0.0),
				251: (251.08076, 0.0),
				252: (252.08431, 0.0),
				253: (253.08688, 0.0),
				254: (254.0906, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element. Belongs to actinoid "
				"series. Eight known isotopes, the most common Bk-247, has a "
				"half-life of 1.4*10^3 years. First produced by Glenn T. Seaborg and "
				"associates in 1949 by bombarding americium-241 with alpha "
				"particles."
				)
		)

Cf = Element(
		98,
		"Cf",
		"Californium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=251.0796,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=1172.0,
		density=15.1,
		eleconfig="[Rn] 5f10 7s2",
		oxistates="4, 3*",
		ionenergy=(6.2817, ),
		isotopes={
				237: (237.06207, 0.0),
				238: (238.06141, 0.0),
				239: (239.06242, 0.0),
				240: (240.0623, 0.0),
				241: (241.06373, 0.0),
				242: (242.0637, 0.0),
				243: (243.06543, 0.0),
				244: (244.066001, 0.0),
				245: (245.068049, 0.0),
				246: (246.0688053, 0.0),
				247: (247.071001, 0.0),
				248: (248.072185, 0.0),
				249: (249.0748535, 0.0),
				250: (250.0764061, 0.0),
				251: (251.079587, 0.0),
				252: (252.081626, 0.0),
				253: (253.085133, 0.0),
				254: (254.087323, 0.0),
				255: (255.09105, 0.0),
				256: (256.09344, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element. Belongs to actinoid "
				"series. Cf-251 has a half life of about 700 years. Nine isotopes "
				"are known. Cf-252 is an intense Neutron source, which makes it an "
				"intense Neutron source and gives it a use in Neutron activation "
				"analysis and a possible use as a radiation source in medicine. "
				"First produced by Glenn T. Seaborg and associates in 1950."
				)
		)

Es = Element(
		99,
		"Es",
		"Einsteinium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=252.083,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=1130.0,
		density=0.0,
		eleconfig="[Rn] 5f11 7s2",
		oxistates="3*",
		ionenergy=(6.42, ),
		isotopes={
				240: (240.06892, 0.0),
				241: (241.06854, 0.0),
				242: (242.06975, 0.0),
				243: (243.06955, 0.0),
				244: (244.07088, 0.0),
				245: (245.07132, 0.0),
				246: (246.0729, 0.0),
				247: (247.07366, 0.0),
				248: (248.07547, 0.0),
				249: (249.07641, 0.0),
				250: (250.07861, 0.0),
				251: (251.079992, 0.0),
				252: (252.08298, 0.0),
				253: (253.0848247, 0.0),
				254: (254.088022, 0.0),
				255: (255.090273, 0.0),
				256: (256.0936, 0.0),
				257: (257.09598, 0.0),
				258: (258.09952, 0.0),
				},
		description=(
				"Appearance is unknown, however it is most probably metallic and "
				"silver or gray in color. Radioactive metallic transuranic element "
				"belonging to the actinoids. Es-254 has the longest half-life of the "
				"eleven known isotopes at 270 days. First identified by Albert "
				"Ghiorso and associates in the debris of the 1952 hydrogen bomb "
				"explosion. In 1961 the first microgram quantities of Es-232 were "
				"separated. While einsteinium never exists naturally, if a "
				"sufficient amount was assembled, it would pose a radiation hazard."
				)
		)

Fm = Element(
		100,
		"Fm",
		"Fermium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=257.0951,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=1800.0,
		density=0.0,
		eleconfig="[Rn] 5f12 7s2",
		oxistates="3*",
		ionenergy=(6.5, ),
		isotopes={
				242: (242.07343, 0.0),
				243: (243.07435, 0.0),
				244: (244.07408, 0.0),
				245: (245.07539, 0.0),
				246: (246.0753, 0.0),
				247: (247.07685, 0.0),
				248: (248.077195, 0.0),
				249: (249.07903, 0.0),
				250: (250.079521, 0.0),
				251: (251.081575, 0.0),
				252: (252.082467, 0.0),
				253: (253.085185, 0.0),
				254: (254.0868542, 0.0),
				255: (255.089962, 0.0),
				256: (256.091773, 0.0),
				257: (257.095105, 0.0),
				258: (258.09708, 0.0),
				259: (259.1006, 0.0),
				260: (260.10268, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element, belongs to the "
				"actinoids. Ten known isotopes, most stable is Fm-257 with a "
				"half-life of 10 days. First identified by Albert Ghiorso and "
				"associates in the debris of the first hydrogen-bomb explosion in "
				"1952."
				)
		)

Md = Element(
		101,
		"Md",
		"Mendelevium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=258.0984,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=1100.0,
		density=0.0,
		eleconfig="[Rn] 5f13 7s2",
		oxistates="3*",
		ionenergy=(6.58, ),
		isotopes={
				245: (245.08083, 0.0),
				246: (246.08189, 0.0),
				247: (247.08164, 0.0),
				248: (248.08282, 0.0),
				249: (249.08301, 0.0),
				250: (250.08442, 0.0),
				251: (251.08484, 0.0),
				252: (252.08656, 0.0),
				253: (253.08728, 0.0),
				254: (254.08966, 0.0),
				255: (255.091083, 0.0),
				256: (256.09406, 0.0),
				257: (257.095541, 0.0),
				258: (258.098431, 0.0),
				259: (259.10051, 0.0),
				260: (260.10365, 0.0),
				261: (261.10572, 0.0),
				262: (262.10887, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element. Belongs to the actinoid "
				"series. Only known isotope, Md-256 has a half-life of 1.3 hours. "
				"First identified by Glenn T. Seaborg, Albert Ghiorso and associates "
				"in 1955. Alternative name Unnilunium has been proposed. Named after "
				"the 'inventor' of the periodic table, Dmitri Mendeleev."
				)
		)

No = Element(
		102,
		"No",
		"Nobelium",
		group=3,
		period=7,
		block='f',
		series=10,
		mass=259.101,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=1100.0,
		density=0.0,
		eleconfig="[Rn] 5f14 7s2",
		oxistates="3, 2*",
		ionenergy=(6.65, ),
		isotopes={
				248: (248.0866, 0.0),
				249: (249.08783, 0.0),
				250: (250.08751, 0.0),
				251: (251.08901, 0.0),
				252: (252.088977, 0.0),
				253: (253.09068, 0.0),
				254: (254.090955, 0.0),
				255: (255.093241, 0.0),
				256: (256.094283, 0.0),
				257: (257.096877, 0.0),
				258: (258.09821, 0.0),
				259: (259.10103, 0.0),
				260: (260.10264, 0.0),
				261: (261.10575, 0.0),
				262: (262.1073, 0.0),
				263: (263.11055, 0.0),
				264: (264.11235, 0.0),
				},
		description=(
				"Radioactive metallic transuranic element, belongs to the "
				"actinoids. Seven known isotopes exist, the most stable being No-254 "
				"with a half-life of 255 seconds. First identified with certainty by "
				"Albert Ghiorso and Glenn T. Seaborg in 1966. Unnilbium has been "
				"proposed as an alternative name."
				)
		)

Lr = Element(
		103,
		"Lr",
		"Lawrencium",
		group=3,
		period=7,
		block='d',
		series=10,
		mass=262.1096,
		eleneg=1.3,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=0.0,
		tmelt=1900.0,
		density=0.0,
		eleconfig="[Rn] 5f14 6d 7s2",
		oxistates="3*",
		ionenergy=(4.9, ),
		isotopes={
				251: (251.09436, 0.0),
				252: (252.09537, 0.0),
				253: (253.09521, 0.0),
				254: (254.09645, 0.0),
				255: (255.09668, 0.0),
				256: (256.09863, 0.0),
				257: (257.09956, 0.0),
				258: (258.10181, 0.0),
				259: (259.1029, 0.0),
				260: (260.1055, 0.0),
				261: (261.10688, 0.0),
				262: (262.10963, 0.0),
				263: (263.11129, 0.0),
				264: (264.11404, 0.0),
				265: (265.11584, 0.0),
				266: (266.11931, 0.0),
				},
		description=(
				"Appearance unknown, however it is most likely silvery-white or "
				"grey and metallic. Lawrencium is a synthetic rare-earth metal. "
				"There are eight known radioisotopes, the most stable being Lr-262 "
				"with a half-life of 3.6 hours. Due to the short half-life of "
				"lawrencium, and its radioactivity, there are no known uses for it. "
				"Identified by Albert Ghiorso in 1961 at Berkeley. It was produced "
				"by bombarding californium with boron ions. The name is temporary "
				"IUPAC nomenclature, the origin of the name comes from Ernest O. "
				"Lawrence, the inventor of the cyclotron. If sufficient amounts of "
				"lawrencium were produced, it would pose a radiation hazard."
				)
		)



#!/usr/bin/env python
#
#  atom.py
"""
Represents an atom in a :class:`~chemistry_tools.pubchem.compound.Compound`.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
from itertools import zip_longest
from typing import Any, Dict, FrozenSet, Optional

# 3rd party
from domdf_python_tools.doctools import prettify_docstrings

# this package
from chemistry_tools.elements import ELEMENTS
from chemistry_tools.pubchem.errors import ResponseParseError

__all__ = ["Atom", "parse_atoms"]


@prettify_docstrings
class Atom:
	"""
	Class to represent an atom in a :class:`~chemistry_tools.pubchem.compound.Compound`.

	:param aid: The Atom ID within the owning Compound.
	:param number: The Atomic number for this atom.
	:param x: The x coordinate for this atom.
	:param y: The y coordinate for this atom.
	:param z: The z coordinate for this atom. Will be :py:obj:`None` in 2D Compound records.
	:param charge: Formal charge on atom.
	"""

	def __init__(
			self,
			aid: int,
			number: int,
			x: Optional[float] = None,
			y: Optional[float] = None,
			z: Optional[float] = None,
			charge: int = 0,
			):
		self.aid: int = aid
		self.number: int = number
		self.x: Optional[float] = x
		self.y: Optional[float] = y
		self.z: Optional[float] = z
		self.charge: int = charge

	def __repr__(self) -> str:
		return f"Atom({self.aid}, {self.element})"

	def __eq__(self, other) -> bool:  # noqa: MAN001
		return (
			isinstance(other, type(self))
			and self.aid == other.aid
			and self.element == other.element
			and self.x == other.x
			and self.y == other.y
			and self.z == other.z
			and self.charge == other.charge
			)  # yapf: disable

	@property
	def element(self) -> str:
		"""
		The element symbol for this atom.
		"""

		return ELEMENTS[self.number].symbol

	def to_dict(self) -> Dict[str, Any]:
		"""
		Return a dictionary containing Atom data.
		"""

		data = {"aid": self.aid, "number": self.number, "element": self.element}

		for coord in {'x', 'y', 'z'}:
			if getattr(self, coord) is not None:
				data[coord] = getattr(self, coord)

		if self.charge != 0:
			data["charge"] = self.charge

		return data

	def set_coordinates(self, x: float, y: float, z: Optional[float] = None) -> None:
		"""
		Set all coordinate dimensions at once.
		"""

		self.x = x
		self.y = y
		self.z = z

	@property
	def coordinate_type(self) -> str:
		"""
		Returns whether this atom has 2D or 3D coordinates.
		"""

		if self.z is None:
			return "2d"
		else:
			return "3d"


def parse_atoms(
		atoms_dict: Dict[str, Any],
		coords_dict: Optional[Dict] = None,
		) -> Dict[FrozenSet[int], Atom]:
	"""
	Parse atoms from the given dictionary.

	:param atoms_dict:
	:param coords_dict:
	"""

	atoms: Dict[FrozenSet[int], Atom] = {}

	# Create atoms
	aids = atoms_dict["aid"]
	elements = atoms_dict["element"]

	if not len(aids) == len(elements):
		raise ResponseParseError("Error parsing atom elements")

	for aid, element in zip(aids, elements):
		atoms[aid] = Atom(aid=aid, number=element)

	# Add coordinates
	if coords_dict:
		coord_ids = coords_dict[0]["aid"]

		xs = coords_dict[0]["conformers"][0]['x']
		ys = coords_dict[0]["conformers"][0]['y']
		zs = coords_dict[0]["conformers"][0].get('z', [])

		if not len(coord_ids) == len(xs) == len(ys) == len(atoms) or (zs and not len(zs) == len(coord_ids)):
			raise ResponseParseError("Error parsing atom coordinates")

		for aid, x, y, z in zip_longest(coord_ids, xs, ys, zs):
			atoms[aid].set_coordinates(x, y, z)

	# Add charges
	if "charge" in atoms_dict:
		for charge in atoms_dict["charge"]:
			atoms[charge["aid"]].charge = charge["value"]

	return atoms



#!/usr/bin/env python3
#
#  description.py
"""
Functions to access the name and description of compounds in the PubChem database.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from typing import Any, Dict, List, Sequence, Union

# this package
from chemistry_tools.pubchem.enums import PubChemNamespace
from chemistry_tools.pubchem.properties import rest_get_properties_json
from chemistry_tools.pubchem.pug_rest import do_rest_get

__all__ = [
		"get_iupac_name",
		"get_description",
		"get_common_name",
		"get_compound_id",
		"rest_get_description",
		"parse_description"
		]


def get_iupac_name(name: str) -> str:
	"""
	Returns the systematic IUPAC name for the compound with the given name.

	:param name:
	"""

	data = rest_get_properties_json(name, PubChemNamespace.name, properties="IUPACName")
	iupac_name = data["PropertyTable"]["Properties"][0]["IUPACName"]
	return str(iupac_name)


def get_description(name: str) -> str:
	"""
	Returns the description compound with the given name.

	:param name:
	"""

	data = rest_get_description(name, PubChemNamespace.name)
	parsed_data = parse_description(data)
	return parsed_data[0]["Description"]


def get_common_name(name: str) -> str:
	"""
	Returns the common name for the compound with the given name.

	:param name:
	"""

	data = rest_get_description(name, PubChemNamespace.name)
	parsed_data = parse_description(data)
	return parsed_data[0]["Title"]


def get_compound_id(name: str) -> str:
	"""
	Returns the compound ID (CID) for the compound with the given name.

	:param name:
	"""

	data = rest_get_description(name, PubChemNamespace.name)
	parsed_data = parse_description(data)
	return parsed_data[0]["CID"]


def rest_get_description(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		namespace: Union[PubChemNamespace, str] = PubChemNamespace.name,
		**kwargs,
		) -> Dict[str, Any]:
	"""
	Obtains the description for the given compound from the PubChem REST API.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param kwargs: Optional arguments that ``json.loads`` takes.

	:raises ValueError: If the response body does not contain valid JSON.

	:return: Parsed JSON data
	"""

	return do_rest_get(namespace, identifier, domain="description").json(**kwargs)


def parse_description(description_data: Dict[str, Any]) -> List[Dict]:
	"""
	Parse raw data from the ``description`` endpoint of the REST API.

	:param description_data:

	:return: A list of dictionaries containing the CID, Title and Description for each compound
	"""

	compounds = {}
	fields = {"Title", "Description"}

	for entry in description_data["InformationList"]["Information"]:

		cid = entry["CID"]

		if cid not in compounds:
			compounds[cid] = {var: None for var in fields}
			compounds[cid]["CID"] = cid

		for var in fields:
			if var in entry:
				compounds[cid][var] = entry[var]

	return list(compounds.values())



#!/usr/bin/env python3
#
#  names.py
"""
Functions for working with IUPAC names for chemicals.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
import re
from typing import Any, Dict, List, Pattern, Sequence, Tuple

# 3rd party
from pandas import DataFrame  # type: ignore[import]

# this package
from chemistry_tools import cached_requests
from chemistry_tools.constants import prefixes
from chemistry_tools.pubchem.errors import HTTP_ERROR_CODES

__all__ = [
		"get_IUPAC_parts",
		"sort_IUPAC_names",
		"get_IUPAC_sort_order",
		"get_sorted_parts",
		"sort_array_by_name",
		"sort_dataframe_by_name",
		"iupac_name_from_cas",
		"cas_from_iupac_name",
		"multiplier_regex",
		"re_strings",
		]

#: Regular expression to match "multiple" prefixes such as **mono-**.
multiplier_regex = re.compile('*'.join([f"({prefix})" for prefix in prefixes.values()]) + '*')

#: List of regular expressions to decompose an IUPAC name.
re_strings: List[Pattern] = [
		re.compile(r"((\d+),?)+(\d+)-"),
		multiplier_regex,
		re.compile(r"nitro"),
		re.compile(r"phenyl"),
		re.compile(r"aniline"),
		re.compile(r"anisole"),
		re.compile(r"benzene"),
		re.compile(r"centralite"),
		re.compile(r"formamide"),
		re.compile(r"glycerine"),
		re.compile(r"nitrate"),
		re.compile(r"glycol"),
		re.compile(r"phthalate"),
		re.compile(r"picrate"),
		re.compile(r"toluene"),
		re.compile(r"methyl"),
		re.compile(r"(?<!m)ethyl"),
		re.compile(r"propyl"),
		re.compile(r"butyl"),
		re.compile(r" "),
		re.compile(r"\("),
		re.compile(r"\)"),
		re.compile(r"hydroxyl"),
		re.compile(r"amin[oe]"),
		re.compile(r"amide"),
		]

_iupac_subs: List[Tuple[Pattern, str]] = [
		# (Regex, replacement)

		# e.g. Bis(2-Nitrophenyl)Amine -> 2,2'-Dinitrophenylamine
		(re.compile(r"^(bis)(\()(\d)(-)(.*)(phenyl)(\))"), r"\3,\3'-Di\5di\6"),
		# e.g. 2-Nitro-N-(4-nitrophenyl)aniline -> 2,4'-Dinitrophenylaniline
		(re.compile(r"^(\d)(-nitro-n-\()(\d)(-nitro)(.*)(\))"), r"\1,\3'-Dinitro-N-\5"),
		(re.compile(r"-?[Nn]-phenylaniline"), "diphenylamine"),
		(re.compile(r"carbanilide"), "-1,3-diphenylurea"),
		(re.compile(r"(glycerol)(-)(\d)(-nitrate)"), r"\3-mononitroglycerin"),
		(re.compile(r"n,n'-"), "1,3-"),
		(re.compile(r"dipicryl"), "hexanitrodiphenyl"),
		(re.compile(r"picryl$"), "-1,3,5-trinitrobenzene"),
		(re.compile(r"picryl"), "-1,3,5-trinitrophenyl"),
		]

_hyphen_digit_hyphen = re.compile(r"(-)(\d)(-)")


def get_IUPAC_parts(string: str) -> List[str]:
	"""
	Splits an IUPAC name for a compound into its constituent parts.

	:param string: The IUPAC name to split.

	:returns: A list of constituent parts.
	"""

	string = string.lower()

	for regex, sub in _iupac_subs:
		string = regex.sub(sub.lower(), string)

	split_points = set()

	for regex in re_strings:
		for match in list(regex.finditer(string.lower())):
			start, end = match.span()
			if start != end:
				split_points.add(start)
				split_points.add(end)

	for match in _hyphen_digit_hyphen.finditer(string.lower()):
		start, end = match.span()
		if start != end:
			split_points.add(start + 1)

	split_points.discard(0)
	split_points_list = sorted(split_points)
	start_point = 0

	string_chars = list(string)
	elements = []
	for point in split_points_list:
		elements.append(''.join(string_chars[start_point:point]))
		start_point = point

	elements.append(''.join(string_chars[start_point:]))

	# Fixups
	fixups = [
			["guani", "di", "ne"],
			]

	for fixup in fixups:
		length = len(fixup)
		for i in range(len(elements)):
			if elements[i:i + length] == fixup:
				elements = elements[:i] + [''.join(fixup)] + elements[i + length:]

	# Remove null elements
	null_elements = {' ', ''}

	elements = [x for x in elements if x not in null_elements]

	while not elements[-1]:
		elements = elements[:-1]

	return elements


#
# from string import ascii_letters
# alphabet = ascii_letters + "0123456789" + ",'" + "- " + '!"#$%&()*+./:;<=>?@[\\]^_`{|}~'


def sort_IUPAC_names(iupac_names: Sequence[str]) -> List[str]:
	"""
	Sort a list of IUPAC names into order.

	:param iupac_names: The list of IUPAC names to sort

	:return: The list of sorted IUPAC names.
	"""

	sort_order = get_IUPAC_sort_order(iupac_names)

	# return [iupac_names[split_names.index(name)] for name in sorted_names]
	return sorted(iupac_names, key=lambda x: sort_order[x])


def get_IUPAC_sort_order(iupac_names: Sequence[str]) -> Dict[str, int]:
	"""
	Returns the order the given IUPAC names should be sorted in.

	Useful when sorting arrays containing data in addition to the name.
	e.g.

	.. code-block:: python

		>>> sort_order = get_IUPAC_sort_order([row[0] for row in data])
		>>> sorted_data = sorted(data, key=lambda row: sort_order[row[0]])

	where row[0] would be the name of the compound

	:param iupac_names: The list of IUPAC names to sort.

	:return: Dictionary mapping the IUPAC names to the order in which they should be sorted.
	"""

	split_names, sorted_names = _get_split_and_sorted_lists(iupac_names)

	sort_order = {}
	for index, name in enumerate(sorted_names):
		sort_order[iupac_names[split_names.index(name)]] = index

	return sort_order


def get_sorted_parts(iupac_names: Sequence[str]) -> List[List[str]]:
	"""
	Returns the constituent parts of the IUPAC names sorted into order.

	The parts returned are in reverse order
	(i.e. ``'diphenylamine'`` becomes ``['amine', 'phenyl', 'di']``).

	:param iupac_names:
	"""

	split_names, sorted_names = _get_split_and_sorted_lists(iupac_names)

	return [split_names[split_names.index(name)] for name in sorted_names]


def _get_split_and_sorted_lists(iupac_names: Sequence[str]) -> Tuple[List[List[str]], List[List[str]]]:
	split_names = []

	for name in iupac_names:
		split_name = get_IUPAC_parts(name.lower())

		if split_name[0].lower() in prefixes.values():
			# no positional information at beginning
			split_name = [' ', *split_name]

		split_names.append(split_name[::-1])

	sorted_names = sorted(split_names)

	return split_names, sorted_names


def sort_array_by_name(array: List[List[Any]], name_col: int = 0, reverse: bool = False) -> List[List[Any]]:
	"""
	Sort a list of lists by the IUPAC name in each row.

	:param array:
	:param name_col: The index of the column containing the IUPAC names
	:param reverse: Whether the names should be sorted in reverse order. Default is :py:obj:`False`, which sorts from A-Z.
	:no-default reverse:

	:return: The sorted array
	"""

	names = [row[name_col] for row in array]
	sort_order = get_IUPAC_sort_order(names)
	sorted_array = sorted(array, key=lambda row: sort_order[row[name_col]], reverse=reverse)
	return sorted_array


def sort_dataframe_by_name(df: DataFrame, name_col: str, reverse: bool = False) -> DataFrame:
	"""
	Sorts a :class:`pandas.DataFrame` by the IUPAC name in each row.

	:param df:
	:param name_col: The name of the column containing the IUPAC names
	:param reverse: Whether the names should be sorted in reverse order. Default is :py:obj:`False`, which sorts from A-Z
	:no-default reverse:

	:return: The sorted :class:`~pandas.DataFrame`
	"""

	names = df[name_col]
	sort_order = get_IUPAC_sort_order(names)
	sorted_df = df.loc[df[name_col].map(sort_order).sort_values(ascending=(not reverse)).index]
	return sorted_df


def iupac_name_from_cas(cas_number: str) -> str:
	"""
	Returns the corresponding IUPAC name for the given CAS registry number.

	:param cas_number: The cas number to search

	:return: The IUPAC name
	"""

	r = cached_requests.get(f"https://cactus.nci.nih.gov/chemical/structure/{cas_number}/iupac_name")
	if r.status_code in HTTP_ERROR_CODES:
		raise ValueError(f"No compound found for CAS registry number {cas_number}.")

	return r.text


def cas_from_iupac_name(iupac_name: str) -> str:
	"""
	Returns the corresponding CAS registry number for the given IUPAC name.

	:param iupac_name: The IUPAC name to search.

	:return: The CAS registry number.
	"""

	r = cached_requests.get(f"https://cactus.nci.nih.gov/chemical/structure/{iupac_name}/cas")
	if r.status_code in HTTP_ERROR_CODES:
		raise ValueError(f"No compound found for name {iupac_name}.")

	return r.text.split('\n')[0]



#!/usr/bin/env python3
#
#  halogens.py
r"""
Group 17: Halogens in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: F

	:class:`~chemistry_tools.elements.classes.Element` representing Fluorine

.. data:: Cl

	:class:`~chemistry_tools.elements.classes.Element` representing Chlorine

.. data:: Br

	:class:`~chemistry_tools.elements.classes.Element` representing Bromine

.. data:: I

	:class:`~chemistry_tools.elements.classes.Element` representing Iodine

.. data:: At

	:class:`~chemistry_tools.elements.classes.Element` representing Astatine

.. data:: Ts

	:class:`~chemistry_tools.elements.classes.Element` representing Tennessine

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

# this package
from .classes import Element

__all__ = ('F', "Cl", "Br", 'I', "At", "Ts")

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

F = Element(
		9,
		'F',
		"Fluorine",
		group=17,
		period=2,
		block='p',
		series=6,
		mass=18.998403163,
		eleneg=3.98,
		eleaffin=3.4011887,
		covrad=0.72,
		atmrad=0.57,
		vdwrad=1.47,
		tboil=85.0,
		tmelt=53.55,
		density=1.58,
		eleconfig="[He] 2s2 2p5",
		oxistates="-1*",
		ionenergy=(
				17.4228,
				34.97,
				62.707,
				87.138,
				114.24,
				157.161,
				185.182,
				953.886,
				1103.089,
				),
		isotopes={
				14: (14.03506, 0.0),
				15: (15.01801, 0.0),
				16: (16.011466, 0.0),
				17: (17.00209524, 0.0),
				18: (18.000938, 0.0),
				19: (18.99840322, 1.0),
				20: (19.99998132, 0.0),
				21: (20.999949, 0.0),
				22: (22.002999, 0.0),
				23: (23.00357, 0.0),
				24: (24.00812, 0.0),
				25: (25.0121, 0.0),
				26: (26.01962, 0.0),
				27: (27.02676, 0.0),
				28: (28.03567, 0.0),
				29: (29.04326, 0.0),
				30: (30.0525, 0.0),
				31: (31.06043, 0.0),
				},
		description=(
				"A poisonous pale yellow gaseous element belonging to group 17 of "
				"the periodic table (The halogens). It is the most chemically "
				"reactive and electronegative element. It is highly dangerous, "
				"causing severe chemical burns on contact with flesh. Fluorine was "
				"identified by Scheele in 1771 and first isolated by Moissan in "
				"1886."
				)
		)

Cl = Element(
		17,
		"Cl",
		"Chlorine",
		group=17,
		period=3,
		block='p',
		series=6,
		mass=35.4529,
		eleneg=3.16,
		eleaffin=3.612724,
		covrad=0.99,
		atmrad=0.97,
		vdwrad=1.75,
		tboil=239.18,
		tmelt=172.17,
		density=2.95,
		eleconfig="[Ne] 3s2 3p5",
		oxistates="7, 5, 3, 1, -1*",
		ionenergy=(
				12.9676,
				23.81,
				39.61,
				53.46,
				67.8,
				98.03,
				114.193,
				348.28,
				400.05,
				455.62,
				529.97,
				591.97,
				656.69,
				749.75,
				809.39,
				3658.425,
				3946.193,
				),
		isotopes={
				28: (28.02851, 0.0),
				29: (29.01411, 0.0),
				30: (30.00477, 0.0),
				31: (30.99241, 0.0),
				32: (31.98569, 0.0),
				33: (32.9774519, 0.0),
				34: (33.97376282, 0.0),
				35: (34.96885268, 0.7576),
				36: (35.96830698, 0.0),
				37: (36.96590259, 0.2424),
				38: (37.96801043, 0.0),
				39: (38.9680082, 0.0),
				40: (39.97042, 0.0),
				41: (40.97068, 0.0),
				42: (41.97325, 0.0),
				43: (42.97405, 0.0),
				44: (43.97828, 0.0),
				45: (44.98029, 0.0),
				46: (45.98421, 0.0),
				47: (46.98871, 0.0),
				48: (47.99495, 0.0),
				49: (49.00032, 0.0),
				50: (50.00784, 0.0),
				51: (51.01449, 0.0),
				},
		description=(
				"Halogen element. Poisonous greenish-yellow gas. Occurs widely in "
				"nature as sodium chloride in seawater. Reacts directly with many "
				"elements and compounds, strong oxidizing agent. Discovered by Karl "
				"Scheele in 1774. Humphrey David confirmed it as an element in 1810."
				)
		)

Br = Element(
		35,
		"Br",
		"Bromine",
		group=17,
		period=4,
		block='p',
		series=6,
		mass=79.9035,
		eleneg=2.96,
		eleaffin=3.363588,
		covrad=1.14,
		atmrad=1.12,
		vdwrad=1.85,
		tboil=331.85,
		tmelt=265.95,
		density=3.14,
		eleconfig="[Ar] 3d10 4s2 4p5",
		oxistates="7, 5, 3, 1, -1*",
		ionenergy=(
				11.8138,
				21.8,
				36.0,
				47.3,
				59.7,
				88.6,
				103.0,
				192.8,
				),
		isotopes={
				67: (66.96479, 0.0),
				68: (67.95852, 0.0),
				69: (68.95011, 0.0),
				70: (69.94479, 0.0),
				71: (70.93874, 0.0),
				72: (71.93664, 0.0),
				73: (72.93169, 0.0),
				74: (73.929891, 0.0),
				75: (74.925776, 0.0),
				76: (75.924541, 0.0),
				77: (76.921379, 0.0),
				78: (77.921146, 0.0),
				79: (78.9183371, 0.5069),
				80: (79.9185293, 0.0),
				81: (80.9162906, 0.4931),
				82: (81.9168041, 0.0),
				83: (82.91518, 0.0),
				84: (83.916479, 0.0),
				85: (84.915608, 0.0),
				86: (85.918798, 0.0),
				87: (86.920711, 0.0),
				88: (87.92407, 0.0),
				89: (88.92639, 0.0),
				90: (89.93063, 0.0),
				91: (90.93397, 0.0),
				92: (91.93926, 0.0),
				93: (92.94305, 0.0),
				94: (93.94868, 0.0),
				95: (94.95287, 0.0),
				96: (95.95853, 0.0),
				97: (96.9628, 0.0),
				},
		description=(
				"Halogen element. Red volatile liquid at room temperature. Its "
				"reactivity is somewhere between chlorine and iodine. Harmful to "
				"human tissue in a liquid state, the vapour irritates eyes and "
				"throat. Discovered in 1826 by Antoine Balard."
				)
		)

I = Element(
		53,
		'I',
		"Iodine",
		group=17,
		period=5,
		block='p',
		series=6,
		mass=126.90447,
		eleneg=2.66,
		eleaffin=3.059038,
		covrad=1.33,
		atmrad=1.32,
		vdwrad=1.98,
		tboil=457.5,
		tmelt=386.7,
		density=4.94,
		eleconfig="[Kr] 4d10 5s2 5p5",
		oxistates="7, 5, 1, -1*",
		ionenergy=(10.4513, 19.131, 33.0),
		isotopes={
				108: (107.94348, 0.0),
				109: (108.93815, 0.0),
				110: (109.93524, 0.0),
				111: (110.93028, 0.0),
				112: (111.92797, 0.0),
				113: (112.92364, 0.0),
				114: (113.92185, 0.0),
				115: (114.91805, 0.0),
				116: (115.91681, 0.0),
				117: (116.91365, 0.0),
				118: (117.913074, 0.0),
				119: (118.91007, 0.0),
				120: (119.910048, 0.0),
				121: (120.907367, 0.0),
				122: (121.907589, 0.0),
				123: (122.905589, 0.0),
				124: (123.9062099, 0.0),
				125: (124.9046302, 0.0),
				126: (125.905624, 0.0),
				127: (126.904473, 1.0),
				128: (127.905809, 0.0),
				129: (128.904988, 0.0),
				130: (129.906674, 0.0),
				131: (130.9061246, 0.0),
				132: (131.907997, 0.0),
				133: (132.907797, 0.0),
				134: (133.909744, 0.0),
				135: (134.910048, 0.0),
				136: (135.91465, 0.0),
				137: (136.917871, 0.0),
				138: (137.92235, 0.0),
				139: (138.9261, 0.0),
				140: (139.931, 0.0),
				141: (140.93503, 0.0),
				142: (141.94018, 0.0),
				143: (142.94456, 0.0),
				144: (143.94999, 0.0),
				},
		description=(
				"Dark violet nonmetallic element, belongs to group 17 of the "
				"periodic table. Insoluble in water. Required as a trace element for "
				"living organisms. One stable isotope, I-127 exists, in addition to "
				"fourteen radioactive isotopes. Chemically the least reactive of the "
				"halogens, and the most electropositive metallic halogen. Discovered "
				"in 1812 by Courtois."
				)
		)

At = Element(
		85,
		"At",
		"Astatine",
		group=17,
		period=6,
		block='p',
		series=6,
		mass=209.9871,
		eleneg=2.2,
		eleaffin=2.8,
		covrad=1.45,
		atmrad=1.43,
		vdwrad=0.0,
		tboil=610.0,
		tmelt=575.0,
		density=0.0,
		eleconfig="[Xe] 4f14 5d10 6s2 6p5",
		oxistates="7, 5, 3, 1, -1*",
		ionenergy=(),
		isotopes={
				193: (192.99984, 0.0),
				194: (193.99873, 0.0),
				195: (194.996268, 0.0),
				196: (195.99579, 0.0),
				197: (196.99319, 0.0),
				198: (197.99284, 0.0),
				199: (198.99053, 0.0),
				200: (199.990351, 0.0),
				201: (200.988417, 0.0),
				202: (201.98863, 0.0),
				203: (202.986942, 0.0),
				204: (203.987251, 0.0),
				205: (204.986074, 0.0),
				206: (205.986667, 0.0),
				207: (206.985784, 0.0),
				208: (207.98659, 0.0),
				209: (208.986173, 0.0),
				210: (209.987148, 0.0),
				211: (210.9874963, 0.0),
				212: (211.990745, 0.0),
				213: (212.992937, 0.0),
				214: (213.996372, 0.0),
				215: (214.998653, 0.0),
				216: (216.002423, 0.0),
				217: (217.004719, 0.0),
				218: (218.008694, 0.0),
				219: (219.011162, 0.0),
				220: (220.01541, 0.0),
				221: (221.01805, 0.0),
				222: (222.02233, 0.0),
				223: (223.02519, 0.0),
				},
		description=(
				"Radioactive halogen element. Occurs naturally from uranium and "
				"thorium decay. At least 20 known isotopes. At-210, the most stable, "
				"has a half-life of 8.3 hours. Synthesized by nuclear bombardment in "
				"1940 by D.R. Corson, K.R. MacKenzie and E. Segre at the University "
				"of California."
				)
		)

Ts = Element(
		117,
		"Ts",
		"Tennessine",
		group=17,
		period=7,
		block='p',
		series=8,
		isotopes={
				291: (291.20656, 0.0),
				292: (292.20755, 0.0),
				}
		)



#!/usr/bin/env python3
#
#  toxnet.py
"""
Read data from National Library of Medicine TOXNET.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from decimal import Decimal
from typing import Any, Dict

# 3rd party
import requests
from bs4 import BeautifulSoup  # type: ignore[import]  # nodep

# this package
from .property_format import *

__all__ = ["toxnet"]


def toxnet(cas: str) -> Dict[str, Any]:
	"""
	Obtain physical properties for the compound with the given CAS number from the ToxNet database.

	:param cas:
	"""

	try:
		base_url = "https://toxnet.nlm.nih.gov"
		origin_url = f"{base_url}/cgi-bin/sis/search2/r?dbs+hsdb:@term+@rn+@rel+{cas}"
		origin_page = requests.get(origin_url)
		origin_soup = BeautifulSoup(origin_page.text, "html.parser")
		# print(origin_url)
		# print(origin_soup.find("a", {"id": "anch_103"}))
		# print(origin_soup.find("input", {"name": "dfield"}))
		data_url = origin_soup.find("input", {"name": "dfield"}).find_next_sibling('a')["href"][:-4] + "cpp"
		# print(data_url)
		data_page = requests.get(base_url + data_url)
		data_soup = BeautifulSoup(data_page.text, "html.parser")
	except AttributeError:
		raise ValueError(f"No Record was found for {cas}")

	physical_properties: Dict[str, Any] = {}

	for prop in data_soup.findAll("h3"):
		prop_name = str(prop).replace("<h3>", '').replace(":</h3>", '')
		prop_value_and_unit = prop.nextSibling.replace('\n', '')

		if prop_name in ["Molecular Formula", "Molecular Weight"]:
			continue

		if prop_name == "Solubilities":
			prop_name = "Solubility"

		# elif prop_name == "Vapor Density":
		# 	physical_properties[prop_name] = prop_value_and_unit.replace(" (Air= 1)",'')
		# else:
		# 	physical_properties[prop_name] = property_format(prop_value_and_unit)

		physical_properties[prop_name] = {}

		# Parse values
		# Temperatures
		if prop_name in ["Boiling Point", "Melting Point"]:
			prop_value = prop_value_and_unit.split(" deg C")[0]
			try:
				physical_properties[prop_name]["Value"] = Decimal(prop_value)
			except:
				physical_properties[prop_name]["Value"] = property_format(prop_value)
			physical_properties[prop_name]["Unit"] = "°C"
			physical_properties[prop_name]["Description"] = None

		# Strings
		elif prop_name in [
				"Color/Form",
				"Odor",
				"Other Chemical/Physical Properties",
				"Solubility",
				"Spectral Properties",
				]:
			physical_properties[prop_name]["Value"] = property_format(prop_value_and_unit)
			physical_properties[prop_name]["Unit"] = None
			physical_properties[prop_name]["Description"] = None

		# Custom Units &c.
		elif prop_name == "Density/Specific Gravity":
			try:
				physical_properties[prop_name]["Value"] = Decimal(prop_value_and_unit)
			except:
				physical_properties[prop_name]["Value"] = property_format(prop_value_and_unit)
			physical_properties[prop_name]["Unit"] = "kg/m<sup>3</sup>"
			physical_properties[prop_name]["Description"] = None
		elif prop_name == "Vapor Density":
			# prop_value = prop_value_and_unit.split(" ")[0]
			physical_properties[prop_name]["Value"] = property_format(prop_value_and_unit.replace(" (Air= 1)", ''))
			physical_properties[prop_name]["Unit"] = "None"
			physical_properties[prop_name]["Description"] = "Air=1"

		# TODO
		elif prop_name == "Dissociation Constants":
			# prop_value_list = prop_value_and_unit.split(" ")
			# prop_value = prop_value_list[2]
			# physical_properties[prop_name]["Value"] = prop_value
			# physical_properties[prop_name]["Unit"] = f"pKa @ {prop_value_list[4]}°C"
			physical_properties[prop_name]["Unit"] = None
			physical_properties[prop_name]["Value"] = property_format(prop_value_and_unit)
			physical_properties[prop_name]["Description"] = None
		elif prop_name == "Heat of Combustion":
			physical_properties[prop_name]["Value"] = property_format(prop_value_and_unit)
			physical_properties[prop_name]["Unit"] = None
			physical_properties[prop_name]["Description"] = None
		# TODO Extract J/KG value from string
		elif prop_name == "Octanol/Water Partition Coefficient":
			# prop_value = prop_value_and_unit.split("log Kow = ")[1]
			# print(prop_value)
			# print(prop_value_and_unit.split("log Kow = "))
			# TODO Fix
			# physical_properties[prop_name]["Value"] = prop_value
			# physical_properties[prop_name]["Unit"] = "log Kow"
			physical_properties[prop_name]["Value"] = property_format(prop_value_and_unit)
			physical_properties[prop_name]["Unit"] = None
			physical_properties[prop_name]["Description"] = None
		elif prop_name == "Surface Tension":
			# prop_value_list = prop_value_and_unit.split(" ")
			# prop_value = prop_value_list[3]
			# print(prop_value_list)
			# physical_properties[prop_name]["Value"] = prop_value
			# physical_properties[prop_name]["Unit"] = f"N/M @ {prop_value_list[6]}°C"
			physical_properties[prop_name]["Value"] = property_format(prop_value_and_unit)
			physical_properties[prop_name]["Unit"] = None
			physical_properties[prop_name]["Description"] = None
		elif prop_name == "Vapor Pressure":
			# prop_value_list = prop_value_and_unit.split(" ")
			# prop_value = prop_value_list[0]
			# physical_properties[prop_name]["Value"] = prop_value
			# physical_properties[prop_name]["Unit"] = f"mm Hg @ {prop_value_list[4]}°C"
			physical_properties[prop_name]["Value"] = property_format(prop_value_and_unit)
			physical_properties[prop_name]["Unit"] = None
			physical_properties[prop_name]["Description"] = None
	# else:
	# 	physical_properties[prop_name] = prop_value_and_unit

	return physical_properties


# Soup CAMEO Link from PubChem page if necessary

if __name__ == "__main__":
	# stdlib
	import pprint

	pprint.pprint(toxnet("122-39-4"))
# pprint.pprint(toxnet("85-98-3")) # No Record
# pprint.pprint(toxnet("71-43-2"))



#!/usr/bin/env python3
#
#  full_record.py
"""
Functions for access the complete set of data held by PubChem for a compound.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

# stdlib
from typing import Dict, List, Sequence, Union

# this package
from chemistry_tools.pubchem.enums import PubChemNamespace
from chemistry_tools.pubchem.properties import _parse_record_property
from chemistry_tools.pubchem.pug_rest import do_rest_get

__all__ = ["parse_full_record", "rest_get_full_record"]


def parse_full_record(record: Dict) -> List[Dict]:
	"""
	Parse the complete PubChem record for a compound.

	:param record:
	"""

	parsed_records = []

	for compound in record["PC_Compounds"]:
		cid = compound["id"]["id"]["cid"]
		# print(cid)

		counts = compound["count"]
		# pprint(counts)

		properties = []
		for prop in compound["props"]:
			prop = _parse_record_property(prop)
			properties.append(prop)

		if "bonds" in compound:
			bonds = compound["bonds"]
		else:
			bonds = {}

		if "charge" in compound:
			charge = compound["charge"]
		else:
			charge = 0

		parsed_records.append(
				dict(
						atoms=compound["atoms"],
						bonds=bonds,
						charge=charge,
						coords=compound["coords"],
						properties=properties,
						cid=cid,
						counts=counts
						)
				)

	return parsed_records


def rest_get_full_record(
		identifier: Union[str, int, Sequence[Union[str, int]]],
		namespace: Union[PubChemNamespace, str] = PubChemNamespace.name,
		record_type: str = "2d",
		**kwargs,
		) -> Dict:
	"""
	Obtains the full record for the given compound from the PubChem REST API.

	:param identifier: Identifiers (e.g. name, CID) for the compound to look up.
		When using the CID namespace data for multiple compounds can be retrieved at once by
		supplying either a comma-separated string or a list.
	:param namespace: The type of identifier to look up. Valid values are in :class:`~.PubChemNamespace`.
	:param record_type:
	:param kwargs: Optional arguments that ``json.loads`` takes.

	:raises ValueError: If the response body does not contain valid JSON.

	:return: Parsed JSON data
	"""

	return do_rest_get(namespace, identifier, record_type=record_type).json(**kwargs)



#!/usr/bin/env python3
#
#  alkali_metals.py
r"""
Group 1: Alkali Metals in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: Li

	:class:`~chemistry_tools.elements.classes.Element` representing Lithium

.. data:: Na

	:class:`~chemistry_tools.elements.classes.Element` representing Sodium

.. data:: K

	:class:`~chemistry_tools.elements.classes.Element` representing Potassium

.. data:: Rb

	:class:`~chemistry_tools.elements.classes.Element` representing Rubidium

.. data:: Cs

	:class:`~chemistry_tools.elements.classes.Element` representing Caesium

.. data:: Fr

	:class:`~chemistry_tools.elements.classes.Element` representing Francium

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

# this package
from .classes import Element

__all__ = ("Li", "Na", 'K', "Rb", "Cs", "Fr")

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

Li = Element(
		3,
		"Li",
		"Lithium",
		group=1,
		period=2,
		block='s',
		series=3,
		mass=6.94,
		eleneg=0.98,
		eleaffin=0.618049,
		covrad=1.23,
		atmrad=2.05,
		vdwrad=1.82,
		tboil=1615.0,
		tmelt=453.7,
		density=0.53,
		eleconfig="[He] 2s",
		oxistates="1*",
		ionenergy=(5.3917, 75.638, 122.451),
		isotopes={
				3: (3.03078, 0.0),
				4: (4.02719, 0.0),
				5: (5.01254, 0.0),
				6: (6.015122795, 0.0759),
				7: (7.01600455, 0.9241),
				8: (8.02248736, 0.0),
				9: (9.0267895, 0.0),
				10: (10.035481, 0.0),
				11: (11.043798, 0.0),
				12: (12.05378, 0.0),
				},
		description=(
				"Socket silvery metal. First member of group 1 of the periodic "
				"table. Lithium salts are used in psychomedicine."
				)
		)

Na = Element(
		11,
		"Na",
		"Sodium",
		group=1,
		period=3,
		block='s',
		series=3,
		mass=22.98976928,
		eleneg=0.93,
		eleaffin=0.547926,
		covrad=1.54,
		atmrad=2.23,
		vdwrad=2.27,
		tboil=1156.0,
		tmelt=371.0,
		density=0.97,
		eleconfig="[Ne] 3s",
		oxistates="1*",
		ionenergy=(
				5.1391,
				47.286,
				71.64,
				98.91,
				138.39,
				172.15,
				208.47,
				264.18,
				299.87,
				1465.091,
				1648.659,
				),
		isotopes={
				18: (18.02597, 0.0),
				19: (19.013877, 0.0),
				20: (20.007351, 0.0),
				21: (20.9976552, 0.0),
				22: (21.9944364, 0.0),
				23: (22.9897692809, 1.0),
				24: (23.99096278, 0.0),
				25: (24.989954, 0.0),
				26: (25.992633, 0.0),
				27: (26.994077, 0.0),
				28: (27.998938, 0.0),
				29: (29.002861, 0.0),
				30: (30.008976, 0.0),
				31: (31.01359, 0.0),
				32: (32.02047, 0.0),
				33: (33.02672, 0.0),
				34: (34.03517, 0.0),
				35: (35.04249, 0.0),
				36: (36.05148, 0.0),
				37: (37.05934, 0.0),
				},
		description=(
				"Soft silvery reactive element belonging to group 1 of the periodic "
				"table (alkali metals). It is highly reactive, oxidizing in air and "
				"reacting violently with water, forcing it to be kept under oil. It "
				"was first isolated by Humphrey Davy in 1807."
				)
		)

K = Element(
		19,
		'K',
		"Potassium",
		group=1,
		period=4,
		block='s',
		series=3,
		mass=39.0983,
		eleneg=0.82,
		eleaffin=0.501459,
		covrad=2.03,
		atmrad=2.77,
		vdwrad=2.75,
		tboil=1033.0,
		tmelt=336.8,
		density=0.86,
		eleconfig="[Ar] 4s",
		oxistates="1*",
		ionenergy=(
				4.3407,
				31.625,
				45.72,
				60.91,
				82.66,
				100.0,
				117.56,
				154.86,
				175.814,
				503.44,
				564.13,
				629.09,
				714.02,
				787.13,
				861.77,
				968.0,
				1034.0,
				4610.955,
				4933.931,
				),
		isotopes={
				32: (32.02192, 0.0),
				33: (33.00726, 0.0),
				34: (33.99841, 0.0),
				35: (34.98801, 0.0),
				36: (35.981292, 0.0),
				37: (36.97337589, 0.0),
				38: (37.9690812, 0.0),
				39: (38.96370668, 0.932581),
				40: (39.96399848, 0.000117),
				41: (40.96182576, 0.067302),
				42: (41.96240281, 0.0),
				43: (42.960716, 0.0),
				44: (43.96156, 0.0),
				45: (44.960699, 0.0),
				46: (45.961977, 0.0),
				47: (46.961678, 0.0),
				48: (47.965514, 0.0),
				49: (48.96745, 0.0),
				50: (49.97278, 0.0),
				51: (50.97638, 0.0),
				52: (51.98261, 0.0),
				53: (52.98712, 0.0),
				54: (53.9942, 0.0),
				55: (54.99971, 0.0),
				},
		description=(
				"Soft silvery metallic element belonging to group 1 of the periodic "
				"table (alkali metals). Occurs naturally in seawater and a many "
				"minerals. Highly reactive, chemically, it resembles sodium in its "
				"behavior and compounds. Discovered by Sir Humphry Davy in 1807."
				)
		)

Rb = Element(
		37,
		"Rb",
		"Rubidium",
		group=1,
		period=5,
		block='s',
		series=3,
		mass=85.4678,
		eleneg=0.82,
		eleaffin=0.485916,
		covrad=2.16,
		atmrad=2.98,
		vdwrad=0.0,
		tboil=961.0,
		tmelt=312.63,
		density=1.53,
		eleconfig="[Kr] 5s",
		oxistates="1*",
		ionenergy=(
				4.1771,
				27.28,
				40.0,
				52.6,
				71.0,
				84.4,
				99.2,
				136.0,
				150.0,
				277.1,
				),
		isotopes={
				71: (70.96532, 0.0),
				72: (71.95908, 0.0),
				73: (72.95056, 0.0),
				74: (73.944265, 0.0),
				75: (74.93857, 0.0),
				76: (75.9350722, 0.0),
				77: (76.930408, 0.0),
				78: (77.928141, 0.0),
				79: (78.923989, 0.0),
				80: (79.922519, 0.0),
				81: (80.918996, 0.0),
				82: (81.9182086, 0.0),
				83: (82.91511, 0.0),
				84: (83.914385, 0.0),
				85: (84.911789738, 0.7217),
				86: (85.91116742, 0.0),
				87: (86.909180527, 0.2783),
				88: (87.91131559, 0.0),
				89: (88.912278, 0.0),
				90: (89.914802, 0.0),
				91: (90.916537, 0.0),
				92: (91.919729, 0.0),
				93: (92.922042, 0.0),
				94: (93.926405, 0.0),
				95: (94.929303, 0.0),
				96: (95.93427, 0.0),
				97: (96.93735, 0.0),
				98: (97.94179, 0.0),
				99: (98.94538, 0.0),
				100: (99.94987, 0.0),
				101: (100.9532, 0.0),
				102: (101.95887, 0.0),
				},
		description=(
				"Soft silvery metallic element, belongs to group 1 of the periodic "
				"table. Rb-97, the naturally occurring isotope, is radioactive. It "
				"is highly reactive, with properties similar to other elements in "
				"group 1, like igniting spontaneously in air. Discovered "
				"spectroscopically in 1861 by W. Bunsen and G.R. Kirchoff."
				)
		)

Cs = Element(
		55,
		"Cs",
		"Caesium",
		group=1,
		period=6,
		block='s',
		series=3,
		mass=132.90545196,
		eleneg=0.79,
		eleaffin=0.471626,
		covrad=2.35,
		atmrad=3.34,
		vdwrad=0.0,
		tboil=944.0,
		tmelt=301.54,
		density=1.9,
		eleconfig="[Xe] 6s",
		oxistates="1*",
		ionenergy=(3.8939, 25.1),
		isotopes={
				112: (111.9503, 0.0),
				113: (112.94449, 0.0),
				114: (113.94145, 0.0),
				115: (114.93591, 0.0),
				116: (115.93337, 0.0),
				117: (116.92867, 0.0),
				118: (117.926559, 0.0),
				119: (118.922377, 0.0),
				120: (119.920677, 0.0),
				121: (120.917229, 0.0),
				122: (121.91611, 0.0),
				123: (122.912996, 0.0),
				124: (123.912258, 0.0),
				125: (124.909728, 0.0),
				126: (125.909452, 0.0),
				127: (126.907418, 0.0),
				128: (127.907749, 0.0),
				129: (128.906064, 0.0),
				130: (129.906709, 0.0),
				131: (130.905464, 0.0),
				132: (131.9064343, 0.0),
				133: (132.905451933, 1.0),
				134: (133.906718475, 0.0),
				135: (134.905977, 0.0),
				136: (135.9073116, 0.0),
				137: (136.9070895, 0.0),
				138: (137.911017, 0.0),
				139: (138.913364, 0.0),
				140: (139.917282, 0.0),
				141: (140.920046, 0.0),
				142: (141.924299, 0.0),
				143: (142.927352, 0.0),
				144: (143.932077, 0.0),
				145: (144.935526, 0.0),
				146: (145.94029, 0.0),
				147: (146.94416, 0.0),
				148: (147.94922, 0.0),
				149: (148.95293, 0.0),
				150: (149.95817, 0.0),
				151: (150.96219, 0.0),
				},
		description=(
				"Soft silvery-white metallic element belonging to group 1 of the "
				"periodic table. One of the three metals which are liquid at room "
				"temperature. Cs-133 is the natural, and only stable, isotope. "
				"Fifteen other radioisotopes exist. Caesium reacts explosively with "
				"cold water, and ice at temperatures above 157K. Caesium hydroxide "
				"is the strongest base known. Caesium is the most electropositive, "
				"most alkaline and has the least ionization potential of all the "
				"elements. Known uses include the basis of atomic clocks, catalyst "
				"for the hydrogenation of some organic compounds, and in "
				"photoelectric cells. Caesium was discovered by Gustav Kirchoff and "
				"Robert Bunsen in Germany in 1860 spectroscopically. Its "
				"identification was based upon the bright blue lines in its "
				"spectrum. The name comes from the latin word caesius, which means "
				"sky blue. Caesium should be considered highly toxic. Some of the "
				"radioisotopes are even more toxic."
				)
		)

Fr = Element(
		87,
		"Fr",
		"Francium",
		group=1,
		period=7,
		block='s',
		series=3,
		mass=223.0197,
		eleneg=0.7,
		eleaffin=0.0,
		covrad=0.0,
		atmrad=0.0,
		vdwrad=0.0,
		tboil=950.0,
		tmelt=300.0,
		density=0.0,
		eleconfig="[Rn] 7s",
		oxistates="1*",
		ionenergy=(4.0727, ),
		isotopes={
				199: (199.00726, 0.0),
				200: (200.00657, 0.0),
				201: (201.00386, 0.0),
				202: (202.00337, 0.0),
				203: (203.000925, 0.0),
				204: (204.000653, 0.0),
				205: (204.998594, 0.0),
				206: (205.99867, 0.0),
				207: (206.99695, 0.0),
				208: (207.99714, 0.0),
				209: (208.995954, 0.0),
				210: (209.996408, 0.0),
				211: (210.995537, 0.0),
				212: (211.996202, 0.0),
				213: (212.996189, 0.0),
				214: (213.998971, 0.0),
				215: (215.000341, 0.0),
				216: (216.003198, 0.0),
				217: (217.004632, 0.0),
				218: (218.007578, 0.0),
				219: (219.009252, 0.0),
				220: (220.012327, 0.0),
				221: (221.014255, 0.0),
				222: (222.017552, 0.0),
				223: (223.0197359, 0.0),
				224: (224.02325, 0.0),
				225: (225.02557, 0.0),
				226: (226.02939, 0.0),
				227: (227.03184, 0.0),
				228: (228.03573, 0.0),
				229: (229.03845, 0.0),
				230: (230.04251, 0.0),
				231: (231.04544, 0.0),
				232: (232.04977, 0.0),
				},
		description=(
				"Radioactive element, belongs to group 1 of the periodic table. "
				"Found in uranium and thorium ores. The 22 known isotopes are all "
				"radioactive, with the most stable being Fr-223. Its existence was "
				"confirmed in 1939 by Marguerite Perey."
				)
		)



#!/usr/bin/env python3
#
#  parser.py
"""
Functions and parsing formulae.

.. latex:vspace:: -10px
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Also based on ChemPy (https://github.com/bjodah/chempy)
#  |  Copyright (c) 2015-2018, Björn Dahlgren
#  |  All rights reserved.
#  |
#  |  Redistribution and use in source and binary forms, with or without modification,
#  |  are permitted provided that the following conditions are met:
#  |
#  |    Redistributions of source code must retain the above copyright notice, this
#  |    list of conditions and the following disclaimer.
#  |
#  |    Redistributions in binary form must reproduce the above copyright notice, this
#  |    list of conditions and the following disclaimer in the documentation and/or
#  |    other materials provided with the distribution.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  |  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  |  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  |  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
#  |  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  |  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  |  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
#  |  ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  |  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  |  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# stdlib
import re
from collections import defaultdict
from functools import lru_cache
from string import ascii_lowercase, ascii_uppercase
from typing import Dict, Iterable, List, Mapping, Optional, Sequence, Union

# 3rd party
import pyparsing  # nodep

# this package
from chemistry_tools.elements import ELEMENTS

# this package
from ._parser_core import _formula_to_parts, _get_charge, _get_leading_integer
from .latex import _latex_mapping

__all__ = ["string_to_composition", "mass_from_composition"]

_atom = r"([A-Z][a-z+]*)(?:\[(\d+)\])?([+-]?\d+)?"
_atom_re = re.compile(_atom)
_formula_re = re.compile(fr"^({_atom})*$")

#: List of the relative atomic masses of each element, in order.
relative_atomic_masses: List[float] = [element.mass for element in ELEMENTS]

#: List of regular expressions for invalid formulae.
invalid_re: List[str]

#: List of regular expressions for matching elements in formulae.
element_re: List[str]

#: List of regular expressions for matching isotopes in formulae.
isotopes_re: List[str]

# Construct regular expression to match all elements, plus D and T
element_re_dict: Dict[str, List[str]] = {}

for element in (ELEMENTS.symbols + ['D', 'T']):
	if len(element) == 1:
		if element in element_re_dict:
			element_re_dict[element].append('?')
		else:
			element_re_dict[element] = ['?']
	else:
		upper, lower = list(element)
		if upper in element_re_dict:
			element_re_dict[upper].append(lower)
		else:
			element_re_dict[upper] = [lower]

invalid_uppers = list(ascii_uppercase)
invalid_lowers = {upper: list(ascii_lowercase) for upper in list(ascii_uppercase)}
invalid_re = []

element_re = []
for upper, lowers in element_re_dict.items():
	lowers = sorted(lowers)

	if lowers == ['?']:
		element_re.append(upper)
		invalid_uppers.remove(upper)
	elif len(lowers) == 1:
		element_re.append(f"{upper}{lowers[0]}")
	elif '?' in lowers:
		lowers.remove('?')
		invalid_uppers.remove(upper)
		for lower in lowers:
			invalid_lowers[upper].remove(lower)
		element_re.append(f"(({upper}[{''.join(lowers)}]?)(?![{''.join(invalid_lowers[upper])}]))(?![a-z])")
		invalid_re.append(f"{upper}[{''.join(lowers)}][{''.join(invalid_lowers[upper])}]+")
		invalid_re.append(f"{upper}[{''.join(invalid_lowers[upper])}]+")
	else:
		invalid_uppers.remove(upper)
		for lower in lowers:
			invalid_lowers[upper].remove(lower)
		element_re.append(f"({upper}[{''.join(lowers)}]?)(?![a-z])")
		invalid_re.append(f"{upper}[{''.join(lowers)}][{''.join(invalid_lowers[upper])}]+")
		invalid_re.append(f"{upper}[{''.join(invalid_lowers[upper])}]+")

# print(invalid_uppers, invalid_lowers)

element_re.sort()

# Isotopes either as [C12], C[12] or [12C]
isotopes_re = []
for elem in element_re:
	isotopes_re.append(rf"\b{elem}\[[0-9]+\]")
	isotopes_re.append(rf"\[{elem}[0-9]+\]")
	isotopes_re.append(rf"\[[0-9]+{elem}\]")


@lru_cache()
def _get_formula_parser() -> pyparsing.Forward:
	"""
	Create a forward pyparsing parser for chemical formulae.

	BNF for simple chemical formula (no nesting)

		integer :: '0'..'9'+
		element :: 'A'..'Z' 'a'..'z'*
		term :: element [integer]
		formula :: term+


	BNF for nested chemical formula

		integer :: '0'..'9'+
		element :: 'A'..'Z' 'a'..'z'*
		term :: (element | '(' formula ')') [integer]
		formula :: term+

	Notes
	-----
	Based on http://stackoverflow.com/a/18555142/790973
	Copyright 2013 Paul McGuire (http://stackoverflow.com/users/165216/paul-mcguire)
	Licensed under CC-BY-SA 3.0.
	"""

	Forward, Group, OneOrMore = pyparsing.Forward, pyparsing.Group, pyparsing.OneOrMore
	Suppress, Word, nums = pyparsing.Suppress, pyparsing.Word, pyparsing.nums

	LPAR, RPAR = map(Suppress, "()")
	integer = Word(nums)

	# add parse action to convert integers to ints, to support doing addition
	# and multiplication at parse time
	integer.setParseAction(lambda t: int(t[0]))

	element = pyparsing.Regex('|'.join(isotopes_re + element_re) + '$')

	# element = pyparsing.Regex(
	# 		r"A[cglmrstu]|B[aehikr]?|C[adeflmorsu]?|D[bsy]|E[rsu]|F[emr]?|G[ade]"
	# 		r"|H[efgos]?|I[nr]?|Kr?|L[airu]|M[dgnot]|N[abdeiop]?|Os?|P[abdmortu]?"
	# 		r"|R[abefghnu]|S[bcegimnr]?|T[abcehilm]|Uu[bhopqst]|U|V|W|Xe|Yb?|Z[nr]")

	# forward declare 'formula' so it can be used in definition of 'term'
	formula = Forward()

	_term_subgroup = (element | Group(LPAR + formula + RPAR)("subgroup"))
	term = Group(_term_subgroup + pyparsing.Optional(integer, default=1)("mult"))

	# add parse actions for parse-time processing

	# parse action to multiply out subgroups
	def multiplyContents(tokens):  # noqa: MAN001,MAN002
		t = tokens[0]
		# if these tokens contain a subgroup, then use multiplier to
		# extend counts of all elements in the subgroup
		if t.subgroup:
			mult = t.mult
			for term in t.subgroup:
				term[1] *= mult
			return t.subgroup

	term.setParseAction(multiplyContents)

	# add parse action to sum up multiple references to the same element
	def sum_by_element(tokens) -> Optional[pyparsing.ParseResults]:  # noqa: MAN001
		elementsList = [t[0] for t in tokens]

		# construct set to see if there are duplicates
		duplicates = len(elementsList) > len(set(elementsList))

		# if there are duplicate element names, sum up by element and
		# return a new nested pyparsing.ParseResults
		if duplicates:
			ctr: Dict = defaultdict(int)
			for t in tokens:
				ctr[t[0]] += t[1]
			return pyparsing.ParseResults([pyparsing.ParseResults([k, v]) for k, v in ctr.items()])
		return None

	# define contents of a formula as one or more terms
	formula << OneOrMore(term)  # pylint: disable=expression-not-assigned
	formula.setParseAction(sum_by_element)
	return formula


# def _parse_stoich(stoich) -> Dict[int, Any]:
# 	if stoich == 'e':  # special case, the electron is not an element
# 		return {}

# 	symbols = ELEMENTS.symbols + ['D', 'T']

# 	if re.findall('|'.join(invalid_re), stoich):
# 		raise ValueError(f"Unrecognised formula: {stoich}")

# 	return {symbols.index(k) + 1: n for k, n in _get_formula_parser().parseString(stoich)}


def string_to_composition(
		formula: str,
		prefixes: Optional[Iterable[str]] = None,
		suffixes: Sequence[str] = ("(s)", "(l)", "(g)", "(aq)"),
		) -> Dict[int, int]:
	"""
	Parse composition of formula representing a chemical formula.

	:bold-title:`Examples:`

	.. code-block:: python

		>>> string_to_composition('NH4+') == {0: 1, "H": 4, "N": 1}
		True
		>>> string_to_composition('.NHO-(aq)') == {0: -1, "H": 1, "N": 1, "O": 1}
		True
		>>> string_to_composition('Na2CO3.7H2O') == {"Na": 2, "C": 1, "O": 10, "H": 14}
		True

	.. latex:vspace:: -10px

	:param formula: Chemical formula, e.g. ``'H2O'``, ``'Fe+3'``, ``'Cl-'``
	:param prefixes: Prefixes to ignore, e.g. ``('.', 'alpha-')``
	:param suffixes: Suffixes to ignore.

	:return: The composition, as a dictionary mapping atomic number -> multiplicity.
		"Atomic number" 0 represents net charge.
	"""

	if prefixes is None:
		prefixes = _latex_mapping.keys()

	stoich_tok, chg_tok = _formula_to_parts(formula, prefixes, suffixes)[:2]
	tot_comp = {}
	parts = stoich_tok.split('.')

	for idx, stoich in enumerate(parts):
		if idx == 0:
			m = 1
		else:
			m, stoich = _get_leading_integer(stoich)

		# comp = _parse_stoich(stoich)
		if stoich == 'e':  # special case, the electron is not an element
			pass
		else:
			try:
				if re.findall('|'.join(invalid_re), stoich):
					raise ValueError(f"Unrecognised formula: {formula}")

				comp = _get_formula_parser().parse_string(stoich)
			except pyparsing.ParseException:
				raise ValueError(f"Unrecognised formula: {formula}")

		# for k, v in comp.items():
			k: int
			v: int
			for k, v in comp:
				if k not in tot_comp:
					tot_comp[k] = m * v
				else:
					tot_comp[k] += m * v

	if chg_tok is not None:
		tot_comp[0] = _get_charge(chg_tok)

	return tot_comp


def mass_from_composition(composition: Mapping[Union[str, int], int], charge: int = 0) -> float:
	"""
	Calculates molecular mass, in atomic mass units, from atomic weights.

	.. latex:vspace:: -5px

	.. note:: Atomic number 0 denotes charge or "net electron defficiency"

	.. latex:vspace:: -15px

	:bold-title:`Example:`

	.. code-block:: python

		>>> f'{mass_from_composition({0: -1, "H": 1, 8: 1}):.2f}'
		'17.01'

	:param composition: Mapping of :class:`str` or :class:`int` (element symbol or atomic number) to int (coefficient)
	:param charge: The charge of the composition.
	"""

	if charge and 0 in composition:
		if charge != composition[0]:
			raise ValueError(
					"'charge' can only be specified once, "
					"either as a keyword argument or as the '0' key of 'composition'"
					)

	mass = 0.0

	for k, v in composition.items():
		if k == 0:  # electron
			mass -= v * 5.489e-4
		elif isinstance(k, str):
			mass += v * ELEMENTS[k].mass
		elif isinstance(k, int):
			mass += v * ELEMENTS[k].mass

	return mass



#!/usr/bin/env python3
#
#  triels.py
r"""
Group 13: Triels (or boron group) in the Periodic Table.

.. raw:: latex

	\begin{multicols}{2}

.. data:: B

	:class:`~chemistry_tools.elements.classes.Element` representing Boron

.. data:: Al

	:class:`~chemistry_tools.elements.classes.Element` representing Aluminium

.. data:: Ga

	:class:`~chemistry_tools.elements.classes.Element` representing Gallium

.. data:: In

	:class:`~chemistry_tools.elements.classes.Element` representing Indium

.. data:: Tl

	:class:`~chemistry_tools.elements.classes.Element` representing Thallium

.. data:: Nh

	:class:`~chemistry_tools.elements.classes.Element` representing Nihonium

.. raw:: latex

	\end{multicols}
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on molmass (https://github.com/cgohlke/molmass)
#  |  Copyright (c) 1990-2020, Christoph Gohlke
#  |  All rights reserved.
#  |  Licensed under the BSD 3-Clause License
#  |  Redistribution and use in source and binary forms, with or without
#  |  modification, are permitted provided that the following conditions are met:
#  |
#  |  1. Redistributions of source code must retain the above copyright notice,
#  |     this list of conditions and the following disclaimer.
#  |
#  |  2. Redistributions in binary form must reproduce the above copyright notice,
#  |     this list of conditions and the following disclaimer in the documentation
#  |     and/or other materials provided with the distribution.
#  |
#  |  3. Neither the name of the copyright holder nor the names of its
#  |     contributors may be used to endorse or promote products derived from
#  |     this software without specific prior written permission.
#  |
#  |  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  |  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  |  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  |  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  |  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  |  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  |  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  |  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  |  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  |  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  |  POSSIBILITY OF SUCH DAMAGE.
#  |
#  Also based on Pyteomics (https://github.com/levitsky/pyteomics)
#  |  Copyright (c) 2011-2015, Anton Goloborodko & Lev Levitsky
#  |  Licensed under the Apache License, Version 2.0 (the "License");
#  |  you may not use this file except in compliance with the License.
#  |  You may obtain a copy of the License at
#  |
#  |    http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS,
#  |  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  |  See the License for the specific language governing permissions and
#  |  limitations under the License.
#  |
#  |  See also:
#  |  Goloborodko, A.A.; Levitsky, L.I.; Ivanov, M.V.; and Gorshkov, M.V. (2013)
#  |  "Pyteomics - a Python Framework for Exploratory Data Analysis and Rapid Software
#  |  Prototyping in Proteomics", Journal of The American Society for Mass Spectrometry,
#  |  24(2), 301–304. DOI: `10.1007/s13361-012-0516-6 <http://dx.doi.org/10.1007/s13361-012-0516-6>`_
#  |
#  |  Levitsky, L.I.; Klein, J.; Ivanov, M.V.; and Gorshkov, M.V. (2018)
#  |  "Pyteomics 4.0: five years of development of a Python proteomics framework",
#  |  Journal of Proteome Research.
#  |  DOI: `10.1021/acs.jproteome.8b00717 <http://dx.doi.org/10.1021/acs.jproteome.8b00717>`_
#
#  Isotope data from http://www.nist.gov/pml/data/comp.cfm .

# this package
from .classes import Element

__all__ = ('B', "Al", "Ga", "In", "Tl", "Nh")

# Isotope 0 Key:
# mass of the most abundant isotope and 1.0 abundance.

B = Element(
		5,
		'B',
		"Boron",
		group=13,
		period=2,
		block='p',
		series=5,
		mass=10.811,
		eleneg=2.04,
		eleaffin=0.279723,
		covrad=0.82,
		atmrad=1.17,
		vdwrad=0.0,
		tboil=4275.0,
		tmelt=2365.0,
		density=2.46,
		eleconfig="[He] 2s2 2p",
		oxistates="3*",
		ionenergy=(8.298, 25.154, 37.93, 59.368, 340.217),
		isotopes={
				6: (6.04681, 0.0),
				7: (7.02992, 0.0),
				8: (8.0246072, 0.0),
				9: (9.0133288, 0.0),
				10: (10.012937, 0.199),
				11: (11.0093054, 0.801),
				12: (12.0143521, 0.0),
				13: (13.0177802, 0.0),
				14: (14.025404, 0.0),
				15: (15.031103, 0.0),
				16: (16.03981, 0.0),
				17: (17.04699, 0.0),
				18: (18.05617, 0.0),
				19: (19.06373, 0.0),
				},
		description=(
				"An element of group 13 of the periodic table. There are two "
				"allotropes, amorphous boron is a brown power, but metallic boron is "
				"black. The metallic form is hard (9.3 on Mohs' scale) and a bad "
				"conductor in room temperatures. It is never found free in nature. "
				"Boron-10 is used in nuclear reactor control rods and shields. It "
				"was discovered in 1808 by Sir Humphry Davy and by J.L. Gay-Lussac "
				"and L.J. Thenard."
				)
		)

Al = Element(
		13,
		"Al",
		"Aluminium",
		group=13,
		period=3,
		block='p',
		series=7,
		mass=26.9815385,
		eleneg=1.61,
		eleaffin=0.43283,
		covrad=1.18,
		atmrad=1.82,
		vdwrad=0.0,
		tboil=2740.0,
		tmelt=933.5,
		density=2.7,
		eleconfig="[Ne] 3s2 3p",
		oxistates="3*",
		ionenergy=(
				5.9858,
				18.828,
				28.447,
				119.99,
				153.71,
				190.47,
				241.43,
				284.59,
				330.21,
				398.57,
				442.07,
				2085.983,
				2304.08,
				),
		isotopes={
				21: (21.02804, 0.0),
				22: (22.01952, 0.0),
				23: (23.007267, 0.0),
				24: (23.9999389, 0.0),
				25: (24.9904281, 0.0),
				26: (25.98689169, 0.0),
				27: (26.98153863, 1.0),
				28: (27.98191031, 0.0),
				29: (28.980445, 0.0),
				30: (29.98296, 0.0),
				31: (30.983947, 0.0),
				32: (31.98812, 0.0),
				33: (32.99084, 0.0),
				34: (33.99685, 0.0),
				35: (34.99986, 0.0),
				36: (36.00621, 0.0),
				37: (37.01068, 0.0),
				38: (38.01723, 0.0),
				39: (39.02297, 0.0),
				40: (40.03145, 0.0),
				41: (41.03833, 0.0),
				42: (42.04689, 0.0),
				},
		description=(
				"Silvery-white lustrous metallic element of group 3 of the periodic "
				"table. Highly reactive but protected by a thin transparent layer of "
				"the oxide which quickly forms in air. There are many alloys of "
				"aluminum, as well as a good number of industrial uses. Makes up "
				"8.1% of the Earth's crust, by weight. Isolated in 1825 by H.C. "
				"Oersted."
				)
		)

Ga = Element(
		31,
		"Ga",
		"Gallium",
		group=13,
		period=4,
		block='p',
		series=7,
		mass=69.723,
		eleneg=1.81,
		eleaffin=0.41,
		covrad=1.26,
		atmrad=1.81,
		vdwrad=1.87,
		tboil=2478.0,
		tmelt=302.92,
		density=5.91,
		eleconfig="[Ar] 3d10 4s2 4p",
		oxistates="3*",
		ionenergy=(5.9993, 20.51, 30.71, 64.0),
		isotopes={
				56: (55.99491, 0.0),
				57: (56.98293, 0.0),
				58: (57.97425, 0.0),
				59: (58.96337, 0.0),
				60: (59.95706, 0.0),
				61: (60.94945, 0.0),
				62: (61.944175, 0.0),
				63: (62.9392942, 0.0),
				64: (63.9368387, 0.0),
				65: (64.9327348, 0.0),
				66: (65.931589, 0.0),
				67: (66.9282017, 0.0),
				68: (67.9279801, 0.0),
				69: (68.9255736, 0.60108),
				70: (69.926022, 0.0),
				71: (70.9247013, 0.39892),
				72: (71.9263663, 0.0),
				73: (72.9251747, 0.0),
				74: (73.926946, 0.0),
				75: (74.9265002, 0.0),
				76: (75.9288276, 0.0),
				77: (76.9291543, 0.0),
				78: (77.9316082, 0.0),
				79: (78.93289, 0.0),
				80: (79.93652, 0.0),
				81: (80.93775, 0.0),
				82: (81.94299, 0.0),
				83: (82.94698, 0.0),
				84: (83.95265, 0.0),
				85: (84.957, 0.0),
				86: (85.96312, 0.0),
				},
		description=(
				"Soft silvery metallic element, belongs to group 13 of the periodic "
				"table. The two stable isotopes are Ga-69 and Ga-71. Eight "
				"radioactive isotopes are known, all having short half-lives. "
				"Gallium Arsenide is used as a semiconductor. Corrodes most other "
				"metals by diffusing into their lattice. First identified by "
				"Francois Lecoq de Boisbaudran in 1875."
				)
		)

In = Element(
		49,
		"In",
		"Indium",
		group=13,
		period=5,
		block='p',
		series=7,
		mass=114.818,
		eleneg=1.78,
		eleaffin=0.404,
		covrad=1.44,
		atmrad=2.0,
		vdwrad=1.93,
		tboil=2350.0,
		tmelt=429.78,
		density=7.31,
		eleconfig="[Kr] 4d10 5s2 5p",
		oxistates="3*",
		ionenergy=(5.7864, 18.869, 28.03, 55.45),
		isotopes={
				97: (96.94954, 0.0),
				98: (97.94214, 0.0),
				99: (98.93422, 0.0),
				100: (99.93111, 0.0),
				101: (100.92634, 0.0),
				102: (101.92409, 0.0),
				103: (102.919914, 0.0),
				104: (103.9183, 0.0),
				105: (104.914674, 0.0),
				106: (105.913465, 0.0),
				107: (106.910295, 0.0),
				108: (107.909698, 0.0),
				109: (108.907151, 0.0),
				110: (109.907165, 0.0),
				111: (110.905103, 0.0),
				112: (111.905532, 0.0),
				113: (112.904058, 0.0429),
				114: (113.904914, 0.0),
				115: (114.903878, 0.9571),
				116: (115.90526, 0.0),
				117: (116.904514, 0.0),
				118: (117.906354, 0.0),
				119: (118.905845, 0.0),
				120: (119.90796, 0.0),
				121: (120.907846, 0.0),
				122: (121.91028, 0.0),
				123: (122.910438, 0.0),
				124: (123.91318, 0.0),
				125: (124.9136, 0.0),
				126: (125.91646, 0.0),
				127: (126.91735, 0.0),
				128: (127.92017, 0.0),
				129: (128.9217, 0.0),
				130: (129.92497, 0.0),
				131: (130.92685, 0.0),
				132: (131.93299, 0.0),
				133: (132.93781, 0.0),
				134: (133.94415, 0.0),
				135: (134.94933, 0.0),
				},
		description=(
				"Soft silvery element belonging to group 13 of the periodic table. "
				"The most common natural isotope is In-115, which has a half-life of "
				"6*10^4 years. Five other radioisotopes exist. Discovered in 1863 by "
				"Reich and Richter."
				)
		)

Tl = Element(
		81,
		"Tl",
		"Thallium",
		group=13,
		period=6,
		block='p',
		series=7,
		mass=204.3834,
		eleneg=2.04,
		eleaffin=0.377,
		covrad=1.48,
		atmrad=2.08,
		vdwrad=1.96,
		tboil=1746.0,
		tmelt=577.0,
		density=11.85,
		eleconfig="[Xe] 4f14 5d10 6s2 6p",
		oxistates="3, 1*",
		ionenergy=(6.1082, 20.428, 29.83),
		isotopes={
				176: (176.00059, 0.0),
				177: (176.996427, 0.0),
				178: (177.9949, 0.0),
				179: (178.99109, 0.0),
				180: (179.98991, 0.0),
				181: (180.986257, 0.0),
				182: (181.98567, 0.0),
				183: (182.982193, 0.0),
				184: (183.98187, 0.0),
				185: (184.97879, 0.0),
				186: (185.97833, 0.0),
				187: (186.975906, 0.0),
				188: (187.97601, 0.0),
				189: (188.973588, 0.0),
				190: (189.97388, 0.0),
				191: (190.971786, 0.0),
				192: (191.97223, 0.0),
				193: (192.97067, 0.0),
				194: (193.9712, 0.0),
				195: (194.969774, 0.0),
				196: (195.970481, 0.0),
				197: (196.969575, 0.0),
				198: (197.97048, 0.0),
				199: (198.96988, 0.0),
				200: (199.970963, 0.0),
				201: (200.970819, 0.0),
				202: (201.972106, 0.0),
				203: (202.9723442, 0.2952),
				204: (203.9738635, 0.0),
				205: (204.9744275, 0.7048),
				206: (205.9761103, 0.0),
				207: (206.977419, 0.0),
				208: (207.9820187, 0.0),
				209: (208.985359, 0.0),
				210: (209.990074, 0.0),
				211: (210.99348, 0.0),
				212: (211.99823, 0.0),
				},
		description=(
				"Pure, unreacted thallium appears silvery-white and exhibits a "
				"metallic lustre. Upon reacting with air, it begins to turn "
				"bluish-grey and looks like lead. It is very malleable, and can be "
				"cut with a knife. There are two stable isotopes, and four "
				"radioisotopes, Tl-204 being the most stable with a half-life of "
				"3.78 years. Thallium sulphate was used as a rodenticide. Thallium "
				"sulphine's conductivity changes with exposure to infrared light, "
				"this gives it a use in infrared detectors. Discovered by Sir "
				"William Crookes via spectroscopy. Its name comes from the Greek "
				"word thallos, which means green twig. Thallium and its compounds "
				"are toxic and can cause cancer."
				)
		)

Nh = Element(
		113,
		"Nh",
		"Nihonium",
		group=13,
		period=7,
		block='p',
		series=8,
		isotopes={
				283: (283.17645, 0.0),
				284: (284.17808, 0.0),
				285: (285.17873, 0.0),
				286: (286.18048, 0.0),
				287: (287.18105, 0.0),
				}
		)



#!/usr/bin/env python
#
#  compound.py
"""
Represents a chemical compound.
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#  Based on PubChemPy https://github.com/mcs07/PubChemPy/blob/master/LICENSE
#  |  Copyright 2014 Matt Swain <m.swain@me.com>
#  |  Licensed under the MIT License
#  |
#  |  Permission is hereby granted, free of charge, to any person obtaining a copy
#  |  of this software and associated documentation files (the "Software"), to deal
#  |  in the Software without restriction, including without limitation the rights
#  |  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  |  copies of the Software, and to permit persons to whom the Software is
#  |  furnished to do so, subject to the following conditions:
#
#  |  The above copyright notice and this permission notice shall be included in
#  |  all copies or substantial portions of the Software.
#
#  |  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  |  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  |  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  |  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  |  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  |  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  |  THE SOFTWARE.
#

# stdlib
from typing import Any, Dict, FrozenSet, List, Optional, Sequence, Type, TypeVar, Union

# 3rd party
from domdf_python_tools.bases import Dictable
from domdf_python_tools.doctools import prettify_docstrings
from pandas import DataFrame, Series  # type: ignore[import]

# this package
from chemistry_tools._memoized_property import memoized_property
from chemistry_tools.formulae import Formula
from chemistry_tools.pubchem.atom import Atom, parse_atoms
from chemistry_tools.pubchem.bond import Bond, parse_bonds
from chemistry_tools.pubchem.enums import CoordinateType
from chemistry_tools.pubchem.full_record import parse_full_record, rest_get_full_record
from chemistry_tools.pubchem.properties import (
		force_valid_properties,
		parse_properties,
		rest_get_properties_json,
		valid_properties
		)
from chemistry_tools.pubchem.synonyms import get_synonyms

__all__ = ["Compound", "compounds_to_frame", 'C']

C = TypeVar('C', bound="Compound")


@prettify_docstrings
class Compound(Dictable):
	"""
	Represents a single record from the PubChem Compound database.

	The PubChem Compound database is constructed from the Substance database
	using a standardization and deduplication process.
	Each Compound is uniquely identified by a CID.

	:param title: The title of the compound record (usually the name of the compound)
	:param CID:
	:param description:
	"""

	def __init__(self, title: str, CID: int, description: str, **_):
		super().__init__()

		self.title: str = str(title)
		self.CID: int = int(CID)
		self.description: str = str(description)
		self._properties: Dict = {prop: None for prop in valid_properties}
		self.record_type: str = "2d"
		self._synonyms: Optional[List[str]] = None

		# Pre-cache all properties
		# self.get_properties("all")

		self._has_full_record: bool = False

	@property
	def __dict__(self):  # noqa: MAN002
		return dict(
				title=self.title,
				CID=self.CID,
				description=self.description,
				properties=self._properties,
				record_type=self.record_type,
				counts=self._record["counts"],
				atoms=self._atoms,
				bonds=self._bonds,
				)

	def __repr__(self) -> str:
		return f"Compound({self.cid})" if self.cid else "Compound()"

	def to_series(self) -> Series:
		"""
		Return a pandas :class:`~pandas.Series` containing Compound data.
		"""

		return Series(dict(self))

	@property
	def cid(self) -> int:
		"""
		Returns the ID of this compound.
		"""

		return self.CID

	@property
	def has_full_record(self) -> bool:
		"""
		Returns whether this compound has a full record available.
		"""

		return self._has_full_record

	@memoized_property
	def _record(self) -> Dict[str, Any]:

		# Only requested when required
		record = parse_full_record(rest_get_full_record(self.CID, "cid", self.record_type))[0]
		self._has_full_record = True

		for prop in record["properties"]:
			if not self._properties["CanonicalSMILES"]:
				if prop.label == "SMILES" and prop.name == "Canonical":
					self._properties["CanonicalSMILES"] = prop.value

			if not self._properties["IsomericSMILES"]:
				if prop.label == "SMILES" and prop.name == "Isomeric":
					self._properties["IsomericSMILES"] = prop.value

			if not self._properties["TPSA"]:
				if prop.label == "Topological" and prop.name == "Polar Surface Area":
					self._properties["TPSA"] = prop.value

			if not self._properties["InChIKey"]:
				if prop.label == "InChIKey" and prop.name == "Standard":
					self._properties["InChIKey"] = prop.value

			if not self._properties["InChI"]:
				if prop.label == "InChI" and prop.name == "Standard":
					self._properties["InChI"] = prop.value

			if not self._properties["HBondAcceptorCount"]:
				if prop.label == "Count" and prop.name == "Hydrogen Bond Acceptor":
					self._properties["HBondAcceptorCount"] = prop.value

			if not self._properties["HBondDonorCount"]:
				if prop.label == "Count" and prop.name == "Hydrogen Bond Donor":
					self._properties["HBondDonorCount"] = prop.value

			if not self._properties["RotatableBondCount"]:
				if prop.label == "Count" and prop.name == "Rotatable Bond":
					self._properties["RotatableBondCount"] = prop.value

			if not self._properties["MolecularWeight"]:
				if prop.label == "Molecular Weight" and prop.name is None:
					self._properties["MolecularWeight"] = prop.value

			# TODO: label='Weight', name='MonoIsotopic',
			# label='Molecular Formula', name=None,
			# label='Mass', name='Exact',
			# label='Log P', name='XLogP3'
			#
		return record

	@memoized_property
	def _atoms(self) -> Optional[Dict[FrozenSet[int], Atom]]:
		"""
		Derive Atom objects from the record.
		"""

		if "atoms" not in self._record:
			return None

		atoms_dict = self._record["atoms"]
		coords_dict = self._record.get("coords", None)
		return parse_atoms(atoms_dict, coords_dict)

	@memoized_property
	def _bonds(self) -> Optional[Dict[FrozenSet[int], Bond]]:
		"""
		Derive Bond objects from the record.
		"""

		if "bonds" not in self._record:
			return None

		bonds_dict = self._record["bonds"]
		coords_dict = self._record.get("coords", None)
		return parse_bonds(bonds_dict, coords_dict)

	def precache(self) -> None:
		"""
		Precache all properties for this compound.
		"""

		self.get_properties("all")
		_ = self._atoms
		_ = self._bonds

	@property
	def atoms(self) -> List[Atom]:
		"""
		List of :class:`Atoms <chemistry_tools.pubchem.atom.Atom>` in this Compound.
		"""

		return sorted(self._atoms.values(), key=lambda x: x.aid)

	@property
	def bonds(self) -> List[Bond]:
		"""
		List of :class:`Bonds <chemistry_tools.pubchem.bond.Bond>`
		between :class:`Atoms <chemistry_tools.pubchem.atom.Atom>`
		in this Compound.
		"""  # noqa: D400

		return sorted(self._bonds.values(), key=lambda x: (x.aid1, x.aid2))

	@property
	def coordinate_type(self) -> Optional[str]:
		"""
		The coordinate type of this compound.
		"""
		if CoordinateType.TWO_D in self._record["coords"][0]["type"]:
			return "2d"
		elif CoordinateType.THREE_D in self._record["coords"][0]["type"]:
			return "3d"
		return None

	@property
	def elements(self) -> List[str]:
		"""
		List of element symbols for atoms in this Compound.
		"""

		return [a.element for a in self.atoms]

	def get_properties(self, properties: Union[Sequence[str], str]) -> Dict[str, Any]:
		"""
		Returns the requested properties for the Compound.

		:param properties: The properties to retrieve for the compound.
			Can be either a comma-separated string or a list.
			See :ref:`the table at the start of this chapter <properties table>` for a list of valid properties.

		:return: Dictionary mapping the property names to their values
		"""

		if isinstance(properties, str) and properties.lower() == "all":
			properties = list(valid_properties.keys())

		properties = force_valid_properties(properties)

		cached_properties = []
		properties_to_get = []

		for prop in properties:
			if self._properties[prop] is not None:
				cached_properties.append(prop)
			else:
				properties_to_get.append(prop)

		output = {}

		if properties_to_get:
			# print("Getting from API")
			data = rest_get_properties_json(self.CID, "cid", properties)
			new_properties = parse_properties(data)[0]

			for prop in properties_to_get:
				self._properties[prop] = new_properties[prop]
				output[prop] = new_properties[prop]

		for prop in cached_properties:
			# print("Getting from cache")
			output[prop] = self._properties[prop]

		return output

	def get_property(self, prop: str) -> Any:
		"""
		Get a single property for the compound.

		:param prop: The property to retrieve for the compound.
			See :ref:`the table at the start of this chapter <properties table>` for a list of valid properties.
		"""

		prop = str(prop)

		if prop not in self._properties:
			raise ValueError(f"Unknown property '{prop}'")

		if self._properties[prop] is not None:
			# print("Getting from cache")
			return self._properties[prop]

		else:
			# print("Getting from API")
			data = rest_get_properties_json(self.CID, "cid", prop)
			new_properties = parse_properties(data)[0]

			self._properties[prop] = new_properties[prop]
			return new_properties[prop]

	@property
	def synonyms(self) -> Optional[List[str]]:
		"""
		Returns a list of synonyms for the Compound.
		"""

		if not self._synonyms:
			data = get_synonyms(self.CID, "cid")[0]

			if not data["CID"] == self.CID:
				raise ValueError("Wrong compound returned")

			else:
				self._synonyms = data["synonyms"]

		return self._synonyms

	@classmethod
	def from_cid(cls: Type['C'], cid: Union[str, int], record_type: str = "2d") -> "Compound":
		"""
		Returns the Compound objects for the compound with the given CID.
		"""

		# this package
		from chemistry_tools.pubchem.lookup import get_compounds

		comp = get_compounds(cid, "cid")[0]

		if comp.CID != int(cid):
			raise ValueError("Wrong compound returned")

		comp.record_type = record_type

		return comp

	# Convenience attributes for some properties

	@property
	def molecular_formula(self) -> Formula:
		"""
		Molecular formula.
		"""

		return self.get_property("MolecularFormula")

	@property
	def canonical_smiles(self) -> str:
		"""
		Canonical SMILES, with no stereochemistry information.
		"""

		return self.get_property("CanonicalSMILES")

	smiles = canonical_smiles

	@property
	def charge(self) -> int:
		"""
		The charge of the compound.
		"""

		return self.get_property("Charge")

	@property
	def molecular_weight(self) -> float:
		"""
		Molecular Weight.
		"""

		return float(self.get_property("MolecularWeight"))

	molecular_mass = molecular_weight

	@memoized_property
	def canonicalized(self) -> bool:
		"""
		Whether the compound is canonicalized.
		"""

		for prop in self._record["properties"]:
			if prop.label == "Compound" and prop.name == "Canonicalized":
				return bool(prop.value)

		return False

	def get_iupac_name(self, type_: str = "Systematic") -> Optional[str]:
		r"""
		Return the IUPAC name of this compound.

		:param type\_: The type of IUPAC name.
		"""

		# Allowed, CAS-like Style, Markup, Preferred, Systematic, Traditional
		for prop in self._record["properties"]:
			if prop.label == "IUPAC Name" and prop.name == type_.capitalize():
				return prop.value

		return None

	@memoized_property
	def iupac_name(self) -> Optional[str]:
		"""
		The preferred IUPAC name of this compound.
		"""

		return self.get_iupac_name("Preferred")

	@memoized_property
	def systematic_name(self) -> Optional[str]:
		"""
		The systematic IUPAC name of this compound.
		"""

		return self.get_iupac_name("Systematic")

	@property
	def fingerprint(self) -> Optional[str]:
		"""
		Raw padded and hex-encoded fingerprint, as returned by the PUG REST API.
		"""

		for prop in self._record["properties"]:
			if prop.label == "Fingerprint" and prop.name == "SubStructure Keys":
				return prop.value

		return None

	@property
	def cactvs_fingerprint(self) -> Optional[str]:
		"""
		PubChem CACTVS fingerprint.

		Each bit in the fingerprint represents the presence or absence of one of 881 chemical substructures.

		.. seealso:: ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.txt
		"""  # noqa: D403

		# Skip first 4 bytes (contain length of fingerprint) and last 7 bits (padding) then re-pad to 881 bits

		if self.fingerprint:
			return f"{int(self.fingerprint[8:], 16):020b}"[:-7].zfill(881)
		else:
			return None

	# @memoized_property
	# def hill_formula(self):
	# 	element_count = Counter(self.elements)
	# 	hill = []
	#
	# 	alphabet = sorted(ELEMENTS.symbols)
	#
	# 	if "C" in element_count:
	# 		hill.append("C")
	# 		alphabet.remove("C")
	# 		count = element_count["C"]
	# 		if count > 1:
	# 			hill.append(f"<sub>{count}</sub>")
	# 		if "H" in element_count:
	# 			hill.append("H")
	# 			alphabet.remove("H")
	# 			count = element_count["H"]
	# 			if count > 1:
	# 				hill.append(f"<sub>{count}</sub>")
	#
	# 	for element in alphabet:
	# 		if element in element_count:
	# 			hill.append(element)
	# 			count = element_count[element]
	# 			if count > 1:
	# 				hill.append(f"<sub>{count}</sub>")
	#
	# 	return "".join(hill)


# TODO from record:
# charge
# properties
# label='Compound', name='Canonicalized'
# label='Compound Complexity', name=None
# cid
# counts


# TODO:
def compounds_to_frame(compounds: Union[Compound, List[Compound]]) -> DataFrame:
	"""
	Construct a :class:`~.pandas.DataFrame` from a list of
	:class:`~chemistry_tools.pubchem.compound.Compound` objects.

	:param compounds:
	"""  # noqa: D400

	if isinstance(compounds, Compound):
		compounds = [compounds]

	return DataFrame.from_records([dict(c) for c in compounds], index="CID")
